"""
This script takes candidate mutation table from snakemake pipeline as input and does a lot of downstream analysis
"""
import numpy as np
import pickle
import math
import pandas as pd
import sys
sys.path.insert(0, './miniscripts_for_dNdS')
import os
import re
import copy
import argparse
from scipy import stats

import traceback
# Import Lieberman Lab SNV-calling python package
script_dir = os.path.dirname(os.path.abspath(__file__))
dir_py_scripts = script_dir+"/modules"
sys.path.insert(0, dir_py_scripts)
import snv_module_recoded_with_dNdS as snv
import build_SNP_Tree as bst

parser=argparse.ArgumentParser(prog='Downstream analysis module of AccuSNV',description='SNV calling tool for bacterial isolates using deep learning.')
parser.add_argument('-i','--input_mat',dest='input_mat',type=str,required=True,help="The input mutation table in npz file")
#parser.add_argument('-t','--input_report',dest='input_report',type=str,required=True,help="The input of the full annotation dataframe (snv_table_merge_all_mut_annotations.tsv) output by accusnv_snakemake")
parser.add_argument('-r','--ref_dir',dest='ref_dir',type=str,help="The dir of your reference genomes")

# Manual filtering parameters - used for fill-N modules for individual samples. (Decide whether the base of the specific sample shown as 'N' or the 'ATGC')
parser.add_argument('-c','--min_cov_for_call',dest='min_cov',type=str,help="For the fill-N module: on individual samples, calls must have at least this many fwd+rev reads. Default is 1.")
parser.add_argument('-q','--min_qual_for_call',dest='min_qual',type=str,help="For the fill-N module: on individual samples, calls must have at least this minimum quality score. Default is 30.")
parser.add_argument('-f','--min_freq_for_call',dest='min_freq',type=str,help="For the fill-N module: on individual samples, a call's major allele must have at least this freq. Default is 0.75.")

# Three new manual filtering parameters for further filtering identified SNVs, updated on 2025-05-23 - There parameters are turned off by default.
parser.add_argument('-d','--max_frac_reads_supporting_indel',dest='max_indel',type=float,help="For the manual filtering module: on individual samples, no more than this fraction of reads can support an indel at any given position. (Default: turn off, referece value: -d 0.33)")
parser.add_argument('-g','--max_fraction_ambigious_samples',dest='min_freq',type=float,help="For the manual filtering module: across samples per position, the fraction of samples that can have undefined bases. (Default: turn off, referece value: -g 0.25)")
parser.add_argument('-m','--min_median_coverage_position',dest='min_med_cov',type=int,help="For the manual filtering module:across samples per position, the median coverage. (Default: turn off, referece value: -m 5)")
#parser.add_argument('-p','--max_mean_copynum',dest='max_mean_cp',type=int,help="Mean copy number at a positions across all samples. (Default: turn off, referece value: -p 4)")
#parser.add_argument('-x','--max_max_copynum',dest='max_max_cp',type=int,help="Max maximum copynumber that a site can have across all samples. (Default: turn off, referece value: -x 7)")

parser.add_argument('-b','--exclude_recomb',type=str,help="Whether included SNVs from potential recombinations. Default included. Set \"-b 1\" to exclude these positions in downstream analysis modules.")

parser.add_argument('-o','--output_dir',dest='output_dir',type=str,help="The output dir")
args = parser.parse_args()

def set_para_int(invalue,expect):
    if not invalue:
        invalue=expect
    else:
        invalue=int(invalue)
    return invalue

def set_para_float(invalue,expect):
    if not invalue:
        invalue=expect
    else:
        invalue=float(invalue)
    return invalue

def search_ref_name(refg):
    pre=''
    #fname=''
    for filename in os.listdir(refg):
        if re.search('fa',filename) or re.search('fna',filename):
            pre=re.split('\.',filename)[0]

            break
    return pre

input_mat=args.input_mat
#input_report=args.input_report
ref_dir=args.ref_dir
ref_genome_name = search_ref_name(ref_dir)
fn_min_cov=args.min_cov
fn_min_qual=args.min_qual
fn_min_freq=args.min_freq
max_indel=args.max_indel
min_freq=args.min_freq
min_med_cov=args.min_med_cov


eb=args.exclude_recomb
output_dir=args.output_dir
if not os.path.exists(output_dir):
    os.makedirs(output_dir)

fn_min_cov=set_para_int(fn_min_cov,1)
fn_min_qual=set_para_int(fn_min_qual,30)
fn_min_freq=set_para_float(fn_min_freq,0.75)
eb=set_para_int(eb,0)

#print(max_indel,min_avg_cov,type(max_frac),min_freq,type(min_med_cov),max_mean_cp,max_max_cp)
#exit()

[quals,p,counts,in_outgroup,sample_names,indel_counter] = \
    snv.read_candidate_mutation_table_npz(input_mat)

my_cmt = snv.cmt_data_object( sample_names,
                             in_outgroup,
                             p,
                             counts,
                             quals,
                             indel_counter
                             )

#print(my_cmt.counts)
#exit()
with open(input_mat, 'rb') as f:
    cmt = np.load(f)
    prob = np.array(cmt['prob'])
    label = np.array(cmt['label'])
    recomb = np.array(cmt['recomb'])

# by default - will use label=1 positions
filt_pos= label
filt_pos2 = ~recomb
#print(len(filt_pos))
#print(len(filt_pos2))
#filt_pos=np.array(filt_pos)
#print(my_cmt.p.shape)
#print(filt_pos2)
if eb==1:
    filt_pos=filt_pos & filt_pos2
#print(filt_pos)
#exit()
my_cmt.filter_positions(filt_pos)
#my_cmt.filter_positions(filt_pos2)
my_calls = snv.calls_object( my_cmt )
my_rg = snv.reference_genome_object( ref_dir )
my_rg_annot = my_rg.annotations
#exit()


# Apply looser filters than before (want as many alleles as possible)



'''
Fill-N module: fill N for samples with a loose filter cutoff
'''
# Remove sites within samples that are not high quality
filter_parameter_calls_for_tree = {
                                    'min_cov_for_call' : fn_min_cov, # on individual samples, calls must have at least this many fwd+rev reads
                                    'min_qual_for_call' : fn_min_qual, # on individual samples, calls must have this minimum quality score
                                    'min_major_nt_freq_for_call' : fn_min_freq,  # on individual samples, a call's major allele must have at least this freq
                                    'max_frac_reads_supporting_indel' : max_indel   # on individual samples, no more than this fraction of reads can support an indel at any given position
                                    }

'''
Manual filtering module: filter SNVs using filtering cutoff (same as those in WideVariant)
'''
# Remove sites across samples that are not high quality
filter_parameter_site_across_samples = {
                                        'max_fraction_ambigious_samples' : min_freq, # across samples per position, the fraction of samples that can have undefined bases # default: 1
                                        'min_median_coverage_position' : min_med_cov, # across samples per position, the median coverage # default: 5
                                        }
'''
Fill-N module
'''
my_calls.filter_calls_by_element(
    my_cmt.coverage < filter_parameter_calls_for_tree['min_cov_for_call']
    ) # forward strand coverage too low

my_calls.filter_calls_by_element(
    my_cmt.quals < filter_parameter_calls_for_tree['min_qual_for_call']
    ) # quality too low

my_calls.filter_calls_by_element(
    my_cmt.major_nt_freq < filter_parameter_calls_for_tree['min_major_nt_freq_for_call']
    ) # major allele frequency too low

'''
Manual filtering module
'''

#print(my_cmt.fwd_cov)
#print(filter_parameter_calls_for_tree['min_cov_for_call'])
#exit()

my_calls.filter_calls_by_element(
    my_cmt.fwd_cov < filter_parameter_calls_for_tree['min_cov_for_call']
    ) # forward strand coverage too low

my_calls.filter_calls_by_element(
    my_cmt.rev_cov < filter_parameter_calls_for_tree['min_cov_for_call']
    ) # reverse strand coverage too low

if max_indel:
    with np.errstate(divide='ignore', invalid='ignore'):
        # compute number of reads supporting an indel
        frac_reads_supporting_indel = np.sum(my_cmt.indel_stats,axis=2) / my_cmt.coverage  # sum reads supporting insertion plus reads supporting deletion
        frac_reads_supporting_indel[~np.isfinite(frac_reads_supporting_indel)] = 0
        # note: this fraction can be above zero beacuse the number of reads supporting an indel includes a +/-3 bp window around a given position on the genome
    my_calls.filter_calls_by_element(
        frac_reads_supporting_indel > filter_parameter_calls_for_tree['max_frac_reads_supporting_indel']
    )  # too many reads supporting indels
    print('Indel-based filter finished.')

if min_freq:
    my_calls.filter_calls_by_position(
        my_calls.get_frac_Ns_by_position() > filter_parameter_site_across_samples['max_fraction_ambigious_samples']
    )  # too many samples with ambiuguous calls at this position

if min_med_cov:
    my_calls.filter_calls_by_position(
        np.median(my_cmt.coverage, axis=0) < filter_parameter_site_across_samples['min_median_coverage_position']
    )  # insufficient median coverage across samples at this position


#exit()

#print(my_calls.calls)
#exit()
###############################################

# Filtered calls for outgroup samples only
calls_outgroup = my_calls.get_calls_in_outgroup_only()
# Switch N's (0's) to NaNs
calls_outgroup_N_as_NaN = calls_outgroup.astype('float') # init ()
calls_outgroup_N_as_NaN[ calls_outgroup_N_as_NaN==0 ] = np.nan

# Infer ancestral allele as the most common allele among outgroup samples (could be N)
calls_ancestral = np.zeros( my_calls.num_pos, dtype='int') # init as N's
outgroup_pos_with_calls = np.any(calls_outgroup,axis=0) # positions where the outgroup has calls
calls_ancestral[outgroup_pos_with_calls] = stats.mode( calls_outgroup_N_as_NaN[:,outgroup_pos_with_calls], axis=0, nan_policy='omit' ).mode.squeeze()

#%% Compute mutation quality

# Grab filtered calls from ingroup samples only
calls_ingroup = my_calls.get_calls_in_sample_subset( np.logical_not( my_calls.in_outgroup ) )
quals_ingroup = my_cmt.quals[ np.logical_not( my_calls.in_outgroup ),: ]
num_samples_ingroup = sum( np.logical_not( my_calls.in_outgroup ) )
# Note: Here we are only looking for SNV differences among ingroup samples. If
# you also want to find SNV differences between the ingroup and the outgroup
# samples (eg mutations that have fixed across the ingroup), then you need to
# use calls and quals matrices that include outgroup samples.
#print(calls_ingroup,quals_ingroup)
#exit()
# Compute quality
[ mut_qual, mut_qual_samples ] = snv.compute_mutation_quality( calls_ingroup, quals_ingroup )
# note: returns NaN if there is only one type of non-N call

# Make a table (pandas dataframe) of SNV positions and relevant annotations
# # Pull alleles from reference genome across p
calls_reference = my_rg.get_ref_NTs_as_ints( my_cmt.p )

# # Update ancestral alleles
pos_to_update = ( calls_ancestral==0 )
calls_ancestral[ pos_to_update ] = calls_reference[ pos_to_update ]
#goodpos_bool=arr = np.ones(len(my_cmt.p), dtype=bool)
#goodpos_bool = np.ones(len(my_cmt.p), dtype=bool)
#goodpos_idx = np.where( goodpos_bool )[0]
calls_goodpos_all = my_calls.calls
calls_goodpos_ingroup_all = calls_goodpos_all[ np.logical_not( my_calls.in_outgroup ),: ]

# Filters
filter_SNVs_not_N = ( calls_ingroup != snv.nts2ints('N') ) # mutations must have a basecall (not N)
filter_SNVs_not_ancestral_allele = ( calls_ingroup != np.tile( calls_ancestral, (num_samples_ingroup,1) ) ) # mutations must differ from the ancestral allele

# Fixed mutations per sample per position
fixedmutation = \
    filter_SNVs_not_N \
    & filter_SNVs_not_ancestral_allele 

goodpos_bool = np.any( fixedmutation, axis=0 )
my_cmt.filter_positions( goodpos_bool )
my_calls.filter_positions( goodpos_bool )

goodpos_idx = np.where( goodpos_bool )[0]
#print(goodpos_bool)
#exit()
p_goodpos_all = my_calls.p
goodpos_idx_all = np.where( goodpos_bool)[0]

# Parameters
promotersize = 250; # how far upstream of the nearest gene to annotate something a promoter mutation (not used if no annotation)
mutations_annotated = snv.annotate_mutations( \
    my_rg, \
    p_goodpos_all, \
    np.tile( calls_ancestral[goodpos_idx], (my_cmt.num_samples,1) ), \
    calls_goodpos_ingroup_all, \
    my_cmt, \
    fixedmutation[:,goodpos_idx_all], \
    mut_qual[:,goodpos_bool].flatten(), \
    promotersize \
    )
'''
Output new annotated text report and the SNV-based phylogenetic tree
'''

# Choose subset of samples or positions to use in the tree by idx
samplestoplot = np.arange(my_cmt.num_samples) # default is to use all samples
num_goodpos_all = len(goodpos_idx_all)
goodpos4tree = np.arange(num_goodpos_all) # default is to use all positions
treesampleNamesLong = my_cmt.sample_names
calls_for_treei = my_calls.calls[np.ix_(samplestoplot, goodpos4tree)]  # numpy broadcasting of row_array and col_array requires np.ix_()
calls_for_tree = snv.ints2nts(calls_for_treei)  # NATCG translation

# Sample names for tree
for i, samplename in enumerate(treesampleNamesLong):
    if not samplename[0].isalpha():
        treesampleNamesLong[i] = 'S' + treesampleNamesLong[
            i]  # sample names are modified to make parsing easier downstream
sampleNamesDnapars = ["{:010d}".format(i) for i in range(my_cmt.num_samples)]

#print(calls_ancestral.shape,calls_ancestral)
#print(calls_ancestral[goodpos_idx].shape,calls_ancestral[goodpos_idx])
# Add inferred ancestor and reference
calls_ancestral_for_tree = np.expand_dims(snv.ints2nts(calls_ancestral[goodpos_idx]), axis=0)
calls_reference_for_tree = np.expand_dims(my_rg.get_ref_NTs(my_calls.p), axis=0)
calls_for_tree_all = np.concatenate((calls_ancestral_for_tree, calls_reference_for_tree, calls_for_tree),axis=0)  # first column now outgroup_nts; outgroup_nts[:, None] to make ndims (2) same for both
sampleNamesDnapars_all = np.append(['Sanc', 'Sref'], sampleNamesDnapars)
treesampleNamesLong_all = np.append(['inferred_ancestor', 'reference_genome'], treesampleNamesLong)

try:
    # Build tree
    snv.generate_tree( \
        calls_for_tree_all.transpose(), \
        treesampleNamesLong_all, \
        sampleNamesDnapars_all, \
        ref_genome_name, \
        output_dir, \
        "snv_tree_" + ref_genome_name, \
        buildTree='PS' \
        )
except Exception as e:
    print('#### error skip #####: something wrong in snv.generate_tree... skip...')
    print(f"Error message: {str(e)}")
    traceback.print_exc()



###############################################
### Output the new SNV table in text format ###
###############################################

output_tsv_filename = output_dir + '/snv_table_mutations_annotations.tsv'
snv.write_mutation_table_as_tsv( \
    my_cmt.p, \
    mut_qual[0, goodpos_idx_all], \
    my_cmt.sample_names, \
    mutations_annotated, \
    calls_for_tree, \
    treesampleNamesLong, \
    output_tsv_filename \
 \
    )

################################################################################
### Identify the number of homoplasic SNVs and rebuild the tree for each SNV ###
################################################################################
try:
    bst.mutationtypes(output_dir + "/snv_tree_genome_latest.nwk.tree",output_dir + '/snv_table_mutations_annotations.tsv',2, output_dir)
except Exception as e:
    print('#### error skip #####: something wrong in bst.mutationtypes... skip...')
    print(f"Error message: {str(e)}")
    traceback.print_exc()

#exit()


"""
@author: Alyssa Mitchell, 2025.03.20

Local code block for integration with AccuSNV downstream analysis module (local) - dN/dS calculation
"""
dir_ref_genome = ref_dir
annotation_full = pd.read_csv(output_dir + '/snv_table_mutations_annotations.tsv',sep='\t')
output_directory = output_dir

if not os.path.exists(output_directory):
    os.makedirs(output_directory)
#%%
# Mutation spectrum
NTs = np.array(['A', 'T', 'C', 'G', 'N'])
mutationmatrix, mut_observed, typecounts, prob_nonsyn = snv.mutation_spectrum_module(annotation_full, NTs)

# Notes:
# * Mutation spectrum tallied for each mutant alelle
# * Mutation type only tallied when there is only one mutant allele
# * Assumes all mutants are true de novo mutants
num_muts_for_empirical_spect = 100
if np.sum(mut_observed) >= num_muts_for_empirical_spect:
    # attempt to get mutation spectrum empirically if there were enough mutations
    mut_spec_prob = mut_observed / np.sum(mut_observed)
else:
    # otherwise assume a uniform mutation spectrum
    mut_spec_prob = np.ones(mut_observed.size) / mut_observed.size
    print('Warning! Assuming uniform mutation spectrum.')

# Expected N/S
probnonsyn_expected = snv.compute_expected_dnds(dir_ref_genome, mut_spec_prob)

# this is going to be a bit off from matlab bc of how it deals with alternate start codons
dnds_expected = probnonsyn_expected / (1 - probnonsyn_expected)  # N/S expected from neutral model

# compute observed N/S for fixed and diverse mutations
    # define gene_nums_of_interest if binning mutations for dN/dS analysis, rather than whole-genome dN/dS
p_nonsyn, CI_nonsyn, num_muts_N, num_muts_S = snv.compute_observed_dnds(annotation_full, gene_nums_of_interest=None)
dnds_observed = p_nonsyn / (1 - p_nonsyn)  # N/S observed
# note that in matlab version, binom(0,0) gives CI of [0,1] even when p=NaN. In python, both are NaN

# dN/dS
# relative to neutral model for this KEGG category
dNdS = dnds_observed / dnds_expected

CI_lower = (CI_nonsyn[0] / (1 - CI_nonsyn[0])) / dnds_expected
try:
    CI_upper = (CI_nonsyn[1] / (1 - CI_nonsyn[1])) / dnds_expected
except ZeroDivisionError:
    CI_upper = np.inf


print('dN/dS =', dNdS)

# save output as binary file using pickle
output_dict = {
    'dNdS': dNdS,
    'CI_lower': CI_lower,
    'CI_upper': CI_upper,
    'num_muts_N': num_muts_N,
    'num_muts_S': num_muts_S,
    'p_nonsyn': p_nonsyn,
    'probnonsyn_expected': probnonsyn_expected
}
'''
with open(output_directory+'/data_dNdS.pickle', 'wb') as f:
    pickle.dump(output_dict, f)
'''
np.savez_compressed(output_directory+'/data_dNdS.npz', **output_dict)




# coding: utf-8

# In[ ]:

import os
from pathlib import Path
import sys
import glob
import subprocess


def read_samplesCSV(spls):
    # reads in samples.csv file, format: Batch, Sample,Alignments,ProviderName,Patient
    hdr_check = ['Batch', 'Sample', 'Alignments', 'ProviderName', 'Patient']
    switch = "on"
    file = open(spls, 'r')
    list_path = []
    list_splID = []
    list_providerNames = []
    list_refG = []
    list_patient = []
    for line in file:
        line = line.strip('\n').split(',')
        # Test Header. Note: Even when header wrong code continues (w/ warning), but first line not read.
        if switch == "on":
            if (line == hdr_check):
                print("Passed CSV header check")
            else:
                Warning("CSV did NOT pass header check! Code continues, but first line ignored")
            switch = "off"
            continue
        # build lists
        list_path.append(line[0])
        list_splID.append(line[1])
        list_refG.append(line[2])
        list_providerNames.append(line[3])
        list_patient.append(line[4])
    return [list_path,list_splID,list_refG,list_providerNames,list_patient] # 2019.02.08, Arolyn: removed set(list_patient) provides only unique subject IDs


def split_samplesCSV(PATH_ls,SAMPLE_ls,REF_Genome_ls,PROVIDER_ls,PATIENTID_ls_all):
    # Added by Arolyn, 2019.02.11
    # takes info extracted form samples.csv; saves each line of samples.csv as sample_info.csv in data/{sampleID}
    for i, sample in enumerate(SAMPLE_ls):
        # get info for this sample
        sample_info_csv_text = PATH_ls[i] + ',' + SAMPLE_ls[i] + ',' + REF_Genome_ls[i] + ',' + PROVIDER_ls[i] + ',' + PATIENTID_ls_all[i]
        #print( sample )
        #print( sample_info_csv_text )
        # make data directory for this sample if it doesn't already exist
        if not(os.path.isdir('data/' + sample)):
            os.makedirs('data/' + sample, exist_ok=True)
        # check to see if this mini csv with sample info already exists
        if os.path.isfile('data/' + sample + '/sample_info.csv'):
            # if so, read file
            old_file = open('data/' + sample + '/sample_info.csv','r')
            old_info = old_file.readline()
            old_file.close()
            # check to see if the existing file is consistent with samples.csv
            if not(old_info == sample_info_csv_text):
                # if not, remove the old file and save sample info in a new file
                #print('Information file must be updated.')  
                os.remove('data/' + sample + '/sample_info.csv')
                f = open('data/' + sample + '/sample_info.csv','w')
                f.write(sample_info_csv_text) 
                f.close()
            #else:
            #print('Information file already updated.')              
        else: # if mini csv with sample info does not already exist
            # save sample info in mini csv
            #print('Information file must be created.')  
            f = open('data/' + sample + '/sample_info.csv','w')
            f.write(sample_info_csv_text) 
            f.close()


def findfastqfile(dr,ID,filename):
    fwd=[]
    rev=[]
    potentialhits_forward=glob.glob(dr + '/' + filename +'/*1.fastq.gz')
    potentialhits_reverse=glob.glob(dr + '/' + filename +'/*2.fastq.gz')
    if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
        fwd=potentialhits_forward[0]
        rev=potentialhits_reverse[0]
    elif len(potentialhits_forward)==0 and len(potentialhits_reverse)==0:
        potentialhits_forward=glob.glob(dr + '/' + filename +'*1.fastq.gz')
        potentialhits_reverse=glob.glob(dr + '/' + filename +'*2.fastq.gz')
        if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
            fwd=potentialhits_forward[0]
            rev=potentialhits_reverse[0]
        elif len(potentialhits_forward)==0 and len(potentialhits_reverse)==0:
            potentialhits_forward=glob.glob(dr + '/' + filename +'*1.fastq')
            potentialhits_reverse=glob.glob(dr + '/' + filename +'*2.fastq')
            if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
                subprocess.run("gzip " + potentialhits_forward[0], shell=True)  
                subprocess.run("gzip " + potentialhits_reverse[0], shell=True)
                fwd=potentialhits_forward[0]+'.gz'
                rev=potentialhits_reverse[0]+'.gz'
            else:
                foldername=glob.glob(dr + '/' + filename + '*')
                if foldername and os.path.isdir(foldername[0]):
                    foldername=foldername[0]
                    potentialhits_forward=glob.glob(foldername + '/*' + filename + '*1*.fastq.gz')
                    potentialhits_reverse=glob.glob(foldername + '/*' + filename + '*2*.fastq.gz')
                    if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
                        fwd=potentialhits_forward[0]
                        rev=potentialhits_reverse[0]
                    elif len(potentialhits_forward)==0 and len(potentialhits_reverse)==0:
                        print(foldername + '/*' + filename + '*2*.fastq.gz')
                        potentialhits_forward=glob.glob(foldername +  '/*' + filename + '*1*.fastq')
                        potentialhits_reverse=glob.glob(foldername + '/*' + filename + '*2*.fastq')
                        if len(potentialhits_forward)==1 and len(potentialhits_reverse)==1:
                            subprocess.run("gzip " + potentialhits_forward[0], shell=True)  
                            subprocess.run("gzip " + potentialhits_reverse[0], shell=True)
                            fwd=potentialhits_forward[0]+'.gz'
                            rev=potentialhits_reverse[0]+'.gz'
    if not(fwd) or not(rev):
        raise ValueError('Either no file or more than 1 file found in ' + dr + 'for ' + ID)
    ##zip fastq files if they aren't already zipped
    subprocess.run("gzip " + fwd, shell=True)   
    subprocess.run("gzip " + rev, shell=True)   
    return [fwd, rev]


def makelink(path,sample,providername):
    #When sample is run on a single lane
    #Provider name can be either a COMPLETE directory name or a file name in batch(called path in this fx)
    [fwd_file, rev_file]=findfastqfile(path,sample, providername)
    subprocess.run('ln -s -T ' + fwd_file + ' data/' + sample + '/R1.fq.gz', shell=True)    
    subprocess.run('ln -s -T ' + rev_file + ' data/' + sample + '/R2.fq.gz', shell=True)    
        



def cp_append_files(paths,sample,providername):
    #When sample is run on multiple lanes with same barcode
    fwd_list=''
    rev_list=''
    for path in paths:
        #Provider name can be either a COMPLETE directory name or a file name in batch(called path in this fx)
        [fwd_file, rev_file]=findfastqfile(path,sample, providername)
        fwd_list=fwd_list+ ' ' +fwd_file
        rev_list=rev_list+ ' ' +rev_file
        print(rev_list)
        print(fwd_list)
    subprocess.run("zcat " + fwd_list + ' | gzip > data/' +  sample + '/R1.fq.gz', shell=True)
    subprocess.run("zcat " + rev_list + ' | gzip > data/' +  sample + '/R2.fq.gz', shell=True)




if __name__ == "__main__":
    if len(sys.argv)<1:
        print('Usage: Reads in a csv file containing information of samples to be processed.')
        print('Copies data links to data/ folder and returns for further processing.')
        print('Example: python read_move_samplesCSV.py samples.csv')
    else:
        file=sys.argv[1]
        [PATH_ls, SAMPLE_ls, REF_Genome_ls, PROVIDER_ls, PATIENTID_ls] = read_samplesCSV(file)  
        for i, path in enumerate(PATH_ls):
            print(i)
            if not(os.path.isfile(SAMPLE_ls[i])):
                os.makedirs('data/' + SAMPLE_ls[i], exist_ok=True)
            paths = path.split(' ')
            if len(paths)>1:
                cp_append_files(paths, SAMPLE_ls[i], PROVIDER_ls[i])
            else:
                makelink(path, SAMPLE_ls[i], PROVIDER_ls[i])




#! /usr/bin/env python
#####################################################################
#generate_kmer_distribution.py creates the kmer distribution file needed for est_abundance.py
#Copyright (C) 2016-2020 Jennifer Lu, jlu26@jhmi.edu

#This file is part of Bracken.

#Bracken is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the license, or
#(at your option) any later version.

#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.

#You should have received a copy of the GNU General Public License
#along with this program; if not, see <http://www.gnu.org/licenses/>.

#####################################################################
#Jennifer Lu, jlu26@jhmi.edu
#01/10/2016
#
#This program evaluates the read distribution of each genome
#and saves the number of reads expected for a given species
#to hit a given taxonomy ID
#
#Note: In this version, error messages have been suppressed. 
#
#Input file: Kraken counts file listing the kraken classification of kmers
#from each genome against the full database. This file must contain the following
#tab-delimited columns:
#   - Identification string for the read
#   - Taxonomy ID of the genome being classified 
#   - Taxonomy ID of the classification kraken assigned to the full read   
#   - Sequence length of the read
#   - Individual Classification Taxonomy IDs for each kmer  
#
#Output file: Kmer distribution file with the following tab-delimited columns:
#   - Classification Taxonomy ID that genomes map to
#   - Genome Taxonomy IDs mapping to that classification (each genome is 
#     space-delimited)
#       - For each genome, the genome taxonomy ID, number of mapped kmers,
#         and number of total kmers are listed, separated by colons ':'
#Methods:
#   - main
#   - parse_single_genome   

import sys, argparse
from time import gmtime
from time import strftime

#parse_single_genome method 
#usage: parses a single line from the input file and extracts relevant information
#input: 
#   - single line from the kraken counts file
#returns:
#   - genome taxonomy ID
#   - total number of kmers for that genome
#   - dictionary: {classification id:number of kmers classified at this level}  
def parse_single_genome(curr_str):
    split_str = curr_str.strip().split('\t')
    #Error check for only one value in the line 
    if len(split_str) < 4:
        return [0,0,0]
    #Get the genome taxonomy ID tested
    genome_taxid = split_str[1]
    #Get which taxonomy ID this read length kmer mapped to
    mapped_id_kmers = {}
    total_kmers = 0
    kmer_distr = split_str[3]
    for kmers in kmer_distr.split():
        #Error check for no mapping/incorrect input file format
        if len(kmers.split(':')) == 1:
            #print("No kmers listed for this mapped_taxid: " + curr_str.strip())
            continue
        [curr_m_id, curr_kmers] = kmers.split(':')
        total_kmers += int(curr_kmers)
        if curr_m_id in mapped_id_kmers:
            mapped_id_kmers[curr_m_id] += int(curr_kmers)
        else:
            mapped_id_kmers[curr_m_id] = int(curr_kmers)
    #Error check for mappings
    if len(mapped_id_kmers) == 0:
        #print("No mappings for this taxid: " + curr_str.strip())
        return [0,0,0]
    #Return if correct
    return [genome_taxid, total_kmers, mapped_id_kmers]

#Main Method
def main():
    #Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', dest='in_file', required=True,
        help='Kraken counts file for each genome mapped to the overall database.')
    parser.add_argument('-o', '--output', dest='output', required=True,
        help='Output file containing each classified taxonomy ID and the \
        kmer distributions of all genomes with this classification.')
    args=parser.parse_args()

    #Start Program
    time = strftime("%m-%d-%Y %H:%M:%S", gmtime())
    sys.stdout.write("PROGRAM START TIME: " + time + '\n')

    #Read in all input lines and create a dictionary mapping genome:mapped taxID:reads
    genome_dict = {}
    genome_dict_totalkmers = {}
    num_genomes = 0
    i_file = open(args.in_file, 'r')
    for line in i_file:
        [genome_taxid, total_kmers, mapped_taxids_kmers] = parse_single_genome(line)
        #No classification - ignore  
        if genome_taxid == 0:
            continue
        #Access genome in dictionary
        if genome_taxid not in genome_dict:
            genome_dict[genome_taxid] = {} 
            genome_dict_totalkmers[genome_taxid] = total_kmers
            num_genomes += 1
        else:
            genome_dict_totalkmers[genome_taxid] += total_kmers
        #Either save new kmer count or add to existing
        for m_taxid in mapped_taxids_kmers:
            if m_taxid not in genome_dict[genome_taxid]:
                genome_dict[genome_taxid][m_taxid] = mapped_taxids_kmers[m_taxid]
            else:
                genome_dict[genome_taxid][m_taxid] += mapped_taxids_kmers[m_taxid]
    i_file.close()
    sys.stdout.write('...' + str(num_genomes) + ' total genomes read from kraken output file\n')

    #Remap dictionary to create mapped_taxids:genome_taxid:kmers
    sys.stdout.write('...creating kmer counts file -- lists the number of kmers of each classification per genome\n')
    mapped_taxids_dict = {}
    for genome in genome_dict:
        for m_taxid in genome_dict[genome]:
            #Create new dictionary if needed
            if m_taxid not in mapped_taxids_dict:
                mapped_taxids_dict[m_taxid] = {}
            #Save number of kmers
            mapped_taxids_dict[m_taxid][genome] = genome_dict[genome][m_taxid]
    
    #Output distributions to file
    sys.stdout.write('...creating kmer distribution file -- lists genomes and kmer counts contributing to each genome\n')
    o_file = open(args.output, 'w')
    o_file.write('mapped_taxid\t' + 'genome_taxids:kmers_mapped:total_genome_kmers\n')
    for m_taxid in mapped_taxids_dict:
        o_file.write(m_taxid + '\t') 
        for genome_taxid in mapped_taxids_dict[m_taxid]:
            o_file.write(genome_taxid + ':' + str(mapped_taxids_dict[m_taxid][genome_taxid]))
            o_file.write(':' + str(genome_dict_totalkmers[genome_taxid]))
            o_file.write(' ')
        o_file.write('\n')
    o_file.close()

    #End of program
    time = strftime("%m-%d-%Y %H:%M:%S", gmtime())
    sys.stdout.write("PROGRAM END TIME: " + time + '\n')

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Submit this clustering script for sbatch to snakemake
"""

import os
import sys
import warnings
import subprocess


jobid = sys.argv[-1]

if "Submitted" in jobid:
    jobid = jobid.split('batch job ')[1]

state = subprocess.run(['sacct','-j',jobid,'-X','--format=State'],stdout=subprocess.PIPE).stdout.decode('utf-8')
state = state.split('\n')[2].strip()

map_state={"PENDING":'running',
           "RUNNING":'running', 
           "SUSPENDED":'running', 
           "REQUEUED":'running',
           "CANCELLED":'failed', 
           "CANCELLED+":'failed',
           "COMPLETING":'running', 
           "COMPLETED":'success', 
           "CONFIGURING":'running', 
           "FAILED":'failed',
           "TIMEOUT":'failed',
           "PREEMPTED":'failed',
           "NODE_FAIL":'failed',
           "REVOKED":'failed',
           "SPECIAL_EXIT":'failed',
           "OUT_OF_ME+":'failed',
           "":'running'}


print(map_state[state])



# -*- coding: utf-8 -*-
"""
---Gathers everything together for candidate_mutation_table---
NOTE: Still reads in many *.mat files etc. Further purging of matlab necessary!
Output:
# path_candidate_mutation_table: where to write
# candidate_mutation_table.mat, ex. results/candidate_mutation_table.mat
---
# Inputs (changed to argparse usage):
     path_to_p_file: where to find all_positions.mat
     path_to_sample_names_file: where to find text file with sample names
         (space delimited)
     path_to_outgroup_boolean_file: where to find text file with outgroup
         booleans (space delimited, 1=outgroup, 0=not)
    path_to_list_of_quals_files: where to find text file with list of
       quals.mat files for each sample (space delimited)
     path_to_list_of_diversity_files: where to find text file with list of
       diversity.mat files for each sample (space delimited)
# Output:
     path_candidate_mutation_table: where to write
     candidate_mutation_table.mat, ex. results/candidate_mutation_table.mat
# Note: All paths should be relative to pwd!
## Version history
     This is adapted from TDL's build_mutation_table_master_smaller_file_size_backup.m
  #   Arolyn, 2018.12.19: This script was written as part of the transition to snakemake.
          It performs the part of the case step that gathers data for
         Quals and counts and saves candidate_mutation_table.mat
  #   Arolyn, 2019.02.12: Added another matlab variable that stores indel
          statistics called 'indel_counter'.
  #   Tami, 2019.12.12: Converted into python and also added ability save coverage data
  #   Felix: 2020.01-04: Continous Debugged and adapted script for streamlined Snakemake implementation.
  #                      Added argparse for proper argument parsing and optional coverage matrix build.
  #   Evan: 2022.02.07: Changed to work with fully python version of lab pipeline
  #   Arolyn, 2022.10.17:
        Fixed coverage matrix generation
        Changed inputs for coverage matrices from booleans to output filenames
        Changed coverage matrices to numpy arrays instead of scipy sparse matrices (because the matrices shouldn't be sparse)
  #   Arolyn, 2022.10.23:
        Fixed bug where indices of indel statistics were not correct (line 173): 8:10 -> 38:40
"""

# %%Testing
# os.chdir("/Users/evanqu/Dropbox (MIT)/Lieberman Lab/Personal lab notebooks/Evan/1-Projects/strainslicer/dev/build_cmt_py")
# path_to_p_file="6-case-temp/allpositions.pickle"
# path_to_sample_names_file="6-case-temp/string_sampleID_names.txt"
# path_to_outgroup_boolean_file="6-case-temp/positions/string_outgroup_bool.txt"
# path_to_list_of_quals_files="6-case-temp/string_qual_mat.txt"
# path_to_list_of_diversity_files="6-case-temp/string_diversity_mat.txt"
# path_to_candidate_mutation_table="test_candidate_mutation_table.pickle.gz"


# main(path_to_p_file, path_to_sample_names_file, path_to_outgroup_boolean_file, path_to_list_of_quals_files, path_to_list_of_diversity_files, path_to_candidate_mutation_table, cov_mat_raw, cov_mat_norm ):

# %%
''' load libraries '''
import numpy as np
import pickle
import scipy.io as sio
import os
import sys, argparse
import gzip
from scipy import sparse

''' positional and optional argument parser'''

parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                 description='''\
                            Gathers everything together for candidate_mutation_table.
                            Optional: Builds coverage matrix (optional w/ double-standardized matrix)
                               ''',
                                 epilog="Questions or comments? --> fkey@mit.edu")
parser.add_argument("-p", dest="allpositions", help="All positions p file (.pickle)", required=True, action='store')
parser.add_argument("-s", dest="sampleNames", help="File with sample names", required=True, action='store')
parser.add_argument("-g", dest="outgroupBool", help="String outgroup bool", required=True, action='store')
parser.add_argument("-q", dest="qualfiles", help="String qual matrix paths", required=True, action='store')
parser.add_argument("-d", dest="divfiles", help="String diversity paths", required=True, action='store')
parser.add_argument("-o", dest="candidate_mutation_table",
                    help="Output candidate mutation table. Py pickle structure (*.pickle.gz)", required=True,
                    action='store')
# parser.add_argument("-c", dest="get_cov", help="Set flag to build raw coverage matrix as sparse csr gzip numpy object (dirname+cov_raw_sparsecsr_mat.npz)",action="store_true", default=False)
# parser.add_argument("-n", dest="get_dbl_norm_cov", help="Set flag to build double normalized coverage matrix as sparse csr gzip numpy object (dirname+cov_norm_sparsecsr_mat.npz)",action="store_true", default=False)
parser.add_argument("-c", dest="cov_mat_raw", help="Output raw coverage matrix as sparse csr gzip numpy object (*.npz)",
                    action='store', default='none')
parser.add_argument("-n", dest="cov_mat_norm",
                    help="Output double normalized coverage matrix as sparse csr gzip numpy object (*.npz)",
                    action='store', default='none')
parser.add_argument("-t", dest="dim", help="Specify the number of statistics (default 8)", type=int, default=40)
args = parser.parse_args()

# %%
'''Functions'''


def main(path_to_p_file, path_to_sample_names_file, path_to_outgroup_boolean_file, path_to_list_of_quals_files,
         path_to_list_of_diversity_files, path_to_candidate_mutation_table, path_to_cov_mat_raw, path_to_cov_mat_norm,
         flag_cov_raw, flag_cov_norm, dim):
    pwd = os.getcwd()

    # p: positions on genome that are candidate SNPs
    print('Processing candidate SNP positions...')
    with open(path_to_p_file, 'rb') as f:
        p = pickle.load(f)
    print('Total number of positions: ' + str(len(p)))

    # SampleNames: list of names of all samples
    print('Processing sample names...')
    fname = pwd + '/' + path_to_sample_names_file
    with open(fname, 'r') as f:
        SampleNames = f.read().splitlines()
    numSamples = len(SampleNames)  # save number of samples
    print('Total number of samples: ' + str(numSamples))

    ## in_outgroup: booleans for whether or not each sample is in the outgroup
    print('Processing outgroup booleans...')

    fname = pwd + '/' + path_to_outgroup_boolean_file
    with open(fname, 'r') as f:
        in_outgroup_str = f.read().splitlines()
    
    in_outgroup = np.asarray([s == '1' for s in in_outgroup_str], dtype=bool).reshape(1, len(in_outgroup_str))

    ## Quals: quality score (relating to sample purity) at each position for all samples
    print('Gathering quality scores at each candidate position...')
    # Import list of directories for where to quals for each sample
    fname = pwd + '/' + path_to_list_of_quals_files
    with open(fname, 'r') as f:
        paths_to_quals_files = f.read().splitlines()
    # Make Quals
    Quals = np.zeros((len(p), numSamples), dtype='int')  # initialize
    for i in range(numSamples):
        print('Loading quals matrix for sample: ' + str(i))
        print('Filename: ' + paths_to_quals_files[i])
        with gzip.open(paths_to_quals_files[i], "rb") as f:
            quals_temp = pickle.load(f)
        quals = quals_temp.flatten()
        Quals[:, i] = quals[p - 1]  # -1 convert position to index

    ## counts: counts for each base from forward and reverse reads at each candidate position for all samples
    print('Gathering counts data at each candidate position...\n')

    # Import list of directories for where to diversity file for each sample
    fname = pwd + '/' + path_to_list_of_diversity_files
    with open(fname, 'r') as f:
        paths_to_diversity_files = f.read().splitlines()
    # Load in first diversity to get some stats
    with gzip.open(paths_to_diversity_files[1], 'rb') as f:
        data = np.array(pickle.load(f))
    size = np.shape(data)
    GenomeLength = size[0]

    # Make counts and coverage at the same time
    counts = np.zeros((dim, len(p), numSamples), dtype='uint')  # initialize
    all_coverage_per_bp = np.zeros((GenomeLength, numSamples), dtype='uint')
    indel_counter = np.zeros((2, len(p), numSamples), dtype='uint')


    for i in range(numSamples):
        print('Loading counts matrix for sample: ' + str(i))
        print('Filename: ' + paths_to_diversity_files[i])
        with gzip.open(paths_to_diversity_files[i], 'rb') as f:
            data = np.array(pickle.load(f))
        counts[:, :, i] = data[p - 1, 0:dim].T  # -1 convert position to index

        if flag_cov_raw:
            np.sum(data[:, 0:dim], axis=1, out=all_coverage_per_bp[:, i])

        indel_counter[:, :, i] = data[p - 1, 38:40].T  # Num reads supporting indels and reads supporting deletions
        # -1 convert position to index


    # Normalize coverage by sample and then position; ignore /0 ; turn resulting inf to 0

    if flag_cov_norm:
        with np.errstate(divide='ignore', invalid='ignore'):
            # 1st normalization
            array_cov_norm = (all_coverage_per_bp - np.mean(all_coverage_per_bp, axis=1, keepdims=True)) / np.std(
                all_coverage_per_bp, axis=1,
                keepdims=True)  # ,keepdims=True maintains 2D array (second dim == 1), necessary for braodcasting
            array_cov_norm[~np.isfinite(array_cov_norm)] = 0

            # 2nd normalization
            array_cov_norm = (array_cov_norm - np.mean(array_cov_norm, axis=0, keepdims=True)) / np.std(array_cov_norm,
                                                                                                        axis=0,
                                                                                                        keepdims=True)  # ,keepdims=True maintains 2D array (second dim == 1), necessary for braodcasting
            array_cov_norm[~np.isfinite(array_cov_norm)] = 0

            # Scale and convert to int to save space
            array_cov_norm_scaled = (np.round(array_cov_norm, 3) * 1000).astype('int64')
            print(array_cov_norm_scaled.dtype)

    # Reshape & save matrices

    Quals = Quals.transpose() #quals: num_samples x num_pos
    p = p.transpose() #p: num_pos
    counts = counts.swapaxes(0,2) #counts: num_samples x num_pos x 8
    in_outgroup = in_outgroup.flatten() #in_outgroup: num_samples 
    SampleNames = SampleNames #sampleNames: num_samples
    indel_counter = indel_counter.swapaxes(0,2) #indel_counter: num_samples x num_pos x 2


    outdir = os.path.dirname(path_to_candidate_mutation_table)
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    if flag_cov_raw:
        print("Saving " + path_to_cov_mat_raw)
        all_coverage_per_bp = all_coverage_per_bp.transpose() #all_coverage_per_bp: num_samples x num_pos
        with open(path_to_cov_mat_raw, 'wb') as f:
            np.savez_compressed(path_to_cov_mat_raw, all_coverage_per_bp=all_coverage_per_bp)

    if flag_cov_norm:
        print("Saving " + path_to_cov_mat_norm)
        array_cov_norm_scaled = array_cov_norm_scaled.transpose() #array_cov_norm_scaled: num_samples x num_pos
        with open(path_to_cov_mat_norm, 'wb') as f:
            np.savez_compressed(path_to_cov_mat_norm, array_cov_norm_scaled=array_cov_norm_scaled)

    CMT = {'sample_names': SampleNames,
           'p': p,
           'counts': counts,
           'quals': Quals,
           'in_outgroup': in_outgroup,
           'indel_counter': indel_counter,
           }

    file_path = path_to_candidate_mutation_table
    print("Saving " + path_to_candidate_mutation_table)
    np.savez_compressed(file_path, **CMT)

    print('DONE')


# %%
if __name__ == "__main__":
    path_to_p_file = args.allpositions
    path_to_sample_names_file = args.sampleNames
    path_to_outgroup_boolean_file = args.outgroupBool
    path_to_list_of_quals_files = args.qualfiles
    path_to_list_of_diversity_files = args.divfiles
    path_to_candidate_mutation_table = args.candidate_mutation_table
    path_to_cov_mat_raw = args.cov_mat_raw
    path_to_cov_mat_norm = args.cov_mat_norm
    dim=args.dim
    if path_to_cov_mat_raw == 'none':
        flag_cov_raw = False
    else:
        flag_cov_raw = True
    if path_to_cov_mat_norm == 'none':
        flag_cov_norm = False
    else:
        flag_cov_norm = True
    main(path_to_p_file, path_to_sample_names_file, path_to_outgroup_boolean_file, path_to_list_of_quals_files,
         path_to_list_of_diversity_files, path_to_candidate_mutation_table, path_to_cov_mat_raw, path_to_cov_mat_norm,
         flag_cov_raw, flag_cov_norm,dim)



# read clade files with sample IDs (input)
# re-build fa file names 
# merge per clade and bgzip
# run spades on newly generated file (output)

# 2019.09.09, Arolyn: changed subdirectory sturcture...
# 2019.09.09, Arolyn: running spades via module add on c3ddb, not from Felix's home directory

import argparse,sys
import subprocess # needed for system calls

''' FUNCTIONS'''

def build_sample_file_list(file):
	''' read clade file with sample IDs and transform to files (based on known structure) '''
	# with open(file, "r") as ins:
	ls_fwd = []
	ls_rev = []
	for line in file:
		#line = line.rstrip('\n')
		ls_fwd.append("1-data_processed/"+line+"/filt1.fq.gz")
		ls_rev.append("1-data_processed/"+line+"/filt2.fq.gz")
	return [ls_fwd,ls_rev]

def merge_fq(ls_sample_file,clade,reads2grab=None):
	''' take all sample-files and merge them using system cat call (rev & fwd) '''
	input_file1_string = ' '.join(ls_sample_file[0])
	input_file2_string = ' '.join(ls_sample_file[1])
	ls1 = ls_sample_file[0]
	ls2 = ls_sample_file[1]
	outfile1 = "0-tmp/in1_spades_"+clade+".fq.gz"
	outfile2 = "0-tmp/in2_spades_"+clade+".fq.gz"
	if reads2grab:
		lines2grab = int(reads2grab)*4
	else:
		reads_smallest_fq = subprocess.run('zgrep -c "^+$" ' + input_file1_string + " | cut -d':' -f 2 | sort -n | head -n 1",shell=True, stdout=subprocess.PIPE)
		lines2grab = int(reads_smallest_fq.stdout)*4
		
	for i in range(len(ls1)):
		subprocess.run("gzip -cd " + ls1[i] + " | head -q -n " + str(lines2grab) + " | gzip >> " + outfile1,shell=True)
		subprocess.run("gzip -cd " + ls2[i] + " | head -q -n " + str(lines2grab) + " | gzip >> " + outfile2,shell=True)
	return [outfile1,outfile2]

def run_spades(ls_merge_fq, clade, threads):
	''' run spades. output to 3-spades/clade_[]/ '''
	outfolder = "Assembly/3-spades/clade_"+clade+"/"
	subprocess.run("mkdir -p " + outfolder , shell=True)
	subprocess.run("spades.py --phred-offset 33 --careful -t " + str(threads) + " -1 " + ls_merge_fq[0] + " -2 " + ls_merge_fq[1] + " -o " + outfolder , shell=True)


''' MAIN '''


if __name__ == "__main__":

	
	''' positional and optional argument parser'''
	
	parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
									 description='''\
		Py Script integrated in de-novo genome assembly snakemake routine.
		Step1: Merge previously created and kraken-validated fq.gz files (each contains 250k)
		Step2: Run SPAdes in careful mode.
		Note: Input file has to follow naming scheme in order to allow the script to extract the clade identifier (see def run_spades : [folder]/[samplesPerSubject]/samples[cladeID]_[string.txt])
									''',
									epilog="Questions or comments? --> fkey@mit.edu")
	parser.add_argument("-i", dest='input', help="Input file per clade including sample-IDs validated by kraken", type=argparse.FileType('rt'))
	parser.add_argument('-t', dest='threads',help="Number of threads",type=int,default=1)
	parser.add_argument('-s', dest='cladeid',help="Clade ID (Snakemake wildcard!)",type=str)
	args = parser.parse_args()
	

	# infile = '3-spades/samplesPerSubject/clade10_samples.txt' # from cammand line
	infile = args.input
	file_names = build_sample_file_list(infile)
	cladeID = args.cladeid 
	outfileLs = merge_fq(file_names,cladeID)
	run_spades(outfileLs,cladeID,threads)
	sys.exit()




import os
from setuptools.command.install import install
from setuptools import setup, find_packages
class PostInstallCommand(install):
    def run(self):
        install.run(self)  
        script_path = os.path.join(self.install_lib, 'accusnv', 'slurm_status_script.py')
        if os.path.exists(script_path):
            os.chmod(script_path, 0o755)  
            print(f"Set execute permissions on {script_path}")

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="accusnv",
    version="1.0.0.2",
    description="High-accuracy SNV calling for bacterial isolates using AccuSNV",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Lieberman Lab",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    package_data={
        'accusnv': [
            'experiment_info.yaml',
            'config.yaml',
            'Snakefile',
            'slurm_status_script.py',
            'CNN_models/**',
            'CNN_models/*',
            'CNN_models/*/**',
            'scripts/**',
            'scripts/*',
            'scripts/*/**',
        ]
    },
    cmdclass={
        'install': PostInstallCommand,  
    },
    entry_points={
        "console_scripts": [
            "accusnv=accusnv.accusnv_main:main",
            "accusnv_snakemake=accusnv.accusnv_snakemake:main",
            "accusnv_downstream=accusnv.downstream:main"
        ]
    },
    install_requires=[
        "numpy",
        "pandas",
        "scipy",
        "matplotlib",
        "statsmodels",
        "torch",
        "biopython==1.78",
        "cutadapt",
        "pulp==2.7.0",
        "tqdm",
        "bcbio-gff==0.6.9"
        #"snakemake==7.32.3"
    ],
    extras_require={},
    python_requires=">=3.9,<3.10",
)



# -*- coding: utf-8 -*-
"""

SUMMARY:

    This script demonstrates how to go from a candidate mutation table and a 
    reference genome to a high-quality SNV table and a parsimony tree.

    This script assumes that your input is whole genome sequencing data from
    isolates that has been processed through the Lieberman Lab's standard 
    Snakemake pipeline.
    
    This script is intended to be run interactively rather than as a single 
    function to enable dynamic adjustment of filters to meet the particulars of 
    your sequencing runs, including sample depth and amount of 
    cross-contamination. In our experience, universal filters do not exist, so 
    manual inspection of filtering is essential to SNV calling.


WARNINGS!

    1. The SNV filters illustrated in this script may not be appropriate for 
    your dataset. It is critical that you use the built-in tools for manual 
    inspection of SNV quality. 
    
    2. This script uses the new version of the Lieberman Lab's SNV module, 
    which as of October 2022 is NOT compatible with old versions. If you add 
    functionality from previous versions, you must ensure that data structure
    and indexing are updated appropriately. See documentation for more 
    information on the standards implemented here.


VERSION HISTORY:

    YEAR.MONTH; Name: Add new revisions here!
    
    2022.10; Arolyn: Major overhaul of main script and python module. Not 
    compatible with previous versions. Introduced classes/methods for candidate 
    mutation table data, basecalls, and reference genomes. Implemented 
    consistency in which python packages are used (for example, all numerical
    arrays are now numpy arrays and heterogeneous arrays are now pandas 
    dataframes) and in indexing of genomes and nucleotides. Added many 
    functions for data visualization and quality control. 
    
    2022.04; Tami, Delphine, Laura: Lieberman Lab Hackathon
    
    Additional notes: This module is based on a previous MATLAB version.


@author: Lieberman Lab at MIT. Authors include: Tami Lieberman, Idan Yelin, 
Felix Key, Arolyn Conwill, A. Delphine Tripp, Evan Qu, Laura Markey, Chris Mancuso

"""


#%%#####################
## SET UP ENVIRONMENT ##
########################

# Import python packages
import sys
import os
import re
import copy
import argparse
import numpy as np
import pandas as pd
from scipy import stats
import datetime, time
import matplotlib.pyplot as plt
import traceback


# Some functions needed for subsequent steps
def search_ref_name(refg):
    pre=''
    #fname=''
    for filename in os.listdir(refg):
        if re.search('fa',filename) or re.search('fna',filename):
            pre=re.split('\.',filename)[0]

            break
    return pre
def remove_same(my_calls_in):
    keep_col = []
    for i in range(my_calls_in.calls.shape[1]):
        unique_nonzero_elements = np.unique(my_calls_in.calls[:, i][my_calls_in.calls[:, i] != 0])
        if len(unique_nonzero_elements) < 2:
            my_calls_in.calls[:, i] = 0
            keep_col.append(False)
        else:
            keep_col.append(True)
    keep_col = np.array(keep_col)
    return keep_col

def is_digit(input_string):
    return input_string.isdigit()

class Tee(object):
    def __init__(self, filename):
        self.file = open(filename, 'w')
        self.stdout = sys.stdout

    def __del__(self):
        self.file.close()

    def write(self, data):
        self.file.write(data)
        self.stdout.write(data)

    def flush(self):
        self.file.flush()
        self.stdout.flush()

def plot_snv_counts_gpt(data_dict, odir=None, title="SNV Counts by Sample", figsize=(10, 6),
                    color='#1f77b4', marker='o', markersize=100,
                    xlabel="Sample Name", ylabel="SNV Count", dpi=400):
    """
    Creates a scatter plot of SNV counts for different samples.

    Parameters:
    -----------
    data_dict : dict
        Dictionary with sample names as keys and SNV counts as values
    odir : str
        Output directory path where the figure will be saved. If None, figure is not saved
    title : str
        Title of the plot
    figsize : tuple
        Figure size (width, height) in inches
    color : str
        Color of the markers
    marker : str
        Marker style
    markersize : int
        Size of the markers
    xlabel : str
        Label for x-axis
    ylabel : str
        Label for y-axis
    dpi : int
        Resolution of the output image in dots per inch

    Returns:
    --------
    fig, ax : tuple
        Matplotlib figure and axis objects
    """
    # Create figure and axis
    fig, ax = plt.subplots(figsize=figsize)

    # Extract sample names and counts
    samples = list(data_dict.keys())
    counts = list(data_dict.values())

    # Create x positions (0, 1, 2, ...)
    x_pos = np.arange(len(samples))

    # Create scatter plot
    ax.scatter(x_pos, counts, s=markersize, c=color, marker=marker, alpha=0.7)

    # Only show x-axis labels if sample count is 20 or less
    if len(samples) <= 20:
        ax.set_xticks(x_pos)
        ax.set_xticklabels(samples, rotation=45, ha='right')
    else:
        ax.set_xticks([])
        ax.set_xticklabels([])

    # Add gridlines for better readability
    ax.grid(True, linestyle='--', alpha=0.7)

    # Add labels and title
    ax.set_xlabel(xlabel, fontsize=12)
    ax.set_ylabel(ylabel, fontsize=12)
    ax.set_title(title, fontsize=14)

    # Adjust y-axis to start from 0
    ymin, ymax = ax.get_ylim()
    ax.set_ylim([0, ymax * 1.1])  # Add 10% padding at the top

    # Add value labels for samples with counts greater than 1000
    ct=0
    for i, count in enumerate(counts):
        if count > 1000 and len(samples)>20:
            ax.annotate(samples[i],
                        (x_pos[i], counts[i]),
                        textcoords="offset points",
                        xytext=(0, -50),
                        ha='center', fontsize=10, color='red',rotation=90)
            ct+=1
            if ct>20:break


    # Adjust layout
    plt.tight_layout()

    fig2,ax2 = plt.subplots(figsize=(figsize[0], figsize[1] * 0.7))
    ax2.hist(counts, bins=20, color=color, alpha=0.7)
    ax2.set_xlabel(ylabel + " Distribution", fontsize=12)
    ax2.set_ylabel("Frequency", fontsize=12)
    ax2.set_title("Histogram of SNV Counts")
    ax2.grid(True, linestyle='--', alpha=0.5)

    plt.tight_layout()


    # Save the figure if output directory is provided
    if odir is not None:
        # Create the directory if it doesn't exist
        os.makedirs(odir, exist_ok=True)

        # Define the output file path
        output_path = os.path.join(odir, "snvs_per_sample.png")


        # Save the figure
        fig.savefig(output_path, dpi=dpi, bbox_inches='tight')
        fig2.savefig(os.path.join(odir, "snvs_histogram_per_sample.png"), dpi=dpi, bbox_inches='tight')
        print(f"Figure saved to: {odir}")

    return fig, ax

# To check SNVs of each input sample
def check_snv(data_file_cmt,odir):
    [quals,p,counts,in_outgroup,sample_names,indel_counter] = \
    snv.read_candidate_mutation_table_npz(data_file_cmt)
    
    #print(in_outgroup)
    #exit()
    if not len(in_outgroup)==len(sample_names):
        in_outgroup=np.array([False] * len(sample_names))
    my_cmt = snv.cmt_data_object( sample_names,
                             in_outgroup,
                             p,
                             counts,
                             quals,
                             indel_counter
                             )
    #print(my_cmt.counts[:,:,:8])
    my_calls = snv.calls_object( my_cmt )
    keep_col = remove_same(my_calls)
    #print(keep_col)
    #exit()
    my_cmt.filter_positions(keep_col)
    my_calls.filter_positions(keep_col)
    #print(my_calls.calls)
    def find_min_freq_elements(column):
        #print(column)
        values, counts = np.unique(column, return_counts=True)
        nonzero_mask = values != 0
        values = values[nonzero_mask]
        counts = counts[nonzero_mask]
        sorted_indices = np.argsort(-counts)
        #print(values,counts,sorted_indices)
        #exit()
        min_count = counts[sorted_indices[1]]
        #print(min_count)
        #print(values[counts == min_count])
        #exit()
        #print(min_count)
        #print(values[counts == min_count])
        #exit()
        res=values[counts == min_count]
        if len(res)>1:
            res=[res[0]]
        return res
    #print(my_calls.calls,my_calls.calls.shape)
    min_freq_elements = np.apply_along_axis(find_min_freq_elements, 0, my_calls.calls)[0]
    array=my_calls.calls
    row_match_counts = np.zeros(array.shape[0], dtype=int)
    for col_idx in range(array.shape[1]):
        col_values = array[:, col_idx]
        min_freq_vals = min_freq_elements[col_idx]

        matches = np.isin(col_values, min_freq_vals)
        row_match_counts += matches
    #print(min_freq_elements,row_match_counts)
    #exit()
    dcs=dict(zip(sample_names,row_match_counts))
    plot_snv_counts_gpt(dcs, odir)
    return dcs


# Import Lieberman Lab SNV-calling python package
script_dir = os.path.dirname(os.path.abspath(__file__))
dir_py_scripts = script_dir+"/modules"
sys.path.insert(0, dir_py_scripts)
import snv_module_recoded_with_dNdS as snv # SNV calling module
import build_SNP_Tree as bst
from . import CNN_pred as cnn


# Get timestamp
ts = time.time() 
timestamp = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d_%H-%M-%S') # used in output files

parser=argparse.ArgumentParser(prog='Local analysis module of AccuSNV',description='Apply filters and CNN to call SNVs for closely related bacterial isolates.')
parser.add_argument('-i','--input_mat',dest='input_mat',type=str,required=True,help="The input mutation table in npz file")
parser.add_argument('-c','--input_cov',dest='input_cov',type=str,help="The input coverage table in npz file")
parser.add_argument('-s','--min_cov_for_filter_sample',dest='min_cov_samp',type=str,help="Before running the CNN model, low-quality samples with more than 45%% of positions having zero aligned reads will be filtered out. (default \"-s 45\") You can adjust this threshold with this parameter; to include all samples, set \"-s 100\".")
parser.add_argument('-v','--min_cov_for_filter_pos',dest='min_cov',type=str,help="For the filter module: on individual samples, calls must have at least this many reads on the fwd/rev strands individually. If many samples have low coverage (e.g. <5), then you can set this parameter to smaller value. (e.g. -v 2). Default is 5.")
parser.add_argument('-e','--excluse_samples',dest='exclude_samp',type=str,help="The names of the samples you want to exclude (e.g. -e S1,S2,S3). If you specify a number, such as \"-e 1000\", any sample with more than 1,000 SNVs will be automatically excluded.")
parser.add_argument('-g','--generate_report',dest='generate_rep',type=str,help="If not generate html report and other related files, set to 0. (default: 1)")
parser.add_argument('-r','--rer',dest='ref_genome',type=str,help="The reference genome")
parser.add_argument('-o','--output_dir',dest='output_dir',type=str,help="The output dir")
#parser.add_argument('-o','--output_file',dest='output_file',type=str,help="The output file")
args=parser.parse_args()
input_mat=args.input_mat
input_cov=args.input_cov
min_cov_samp=args.min_cov_samp
min_cov_filt=args.min_cov
refg=args.ref_genome
odir=args.output_dir
greport=args.generate_rep
exclude_samp=args.exclude_samp
if not min_cov_filt:
    min_cov_filt=5
else:
    min_cov_filt=int(min_cov_filt)

if not min_cov_samp:
    min_cov_samp=45
else:
    min_cov_samp=int(min_cov_samp)

if not exclude_samp:
    exclude_samp=''

if is_digit(exclude_samp):
    exclude_samp=int(exclude_samp)
    
if not greport:
    greport=1
else:
    greport=int(greport)
#Build
if not os.path.exists(odir):
    os.makedirs(odir)



#%%#############################
## DATA IMPORT AND PROCESSING ##
################################


#%% Dataset info
dataset_name= 'Your-InputData'
data_file_cmt=input_mat
data_file_cov = input_cov
dir_ref_genome = refg
ref_genome_name = search_ref_name(refg)
#dir_ref_genome = refg+'/'+fname
#samples_to_exclude = ["P-15_O-Ec_S-STOOL_C-C1_D-4","P-15_O-Ec_S-STOOL_C-D3_D-4"] # option to exclude specific samples manually
#samples_to_exclude=["P-13_O-Kp_S-BLOOD_C-H1_D-0"]

dcs=check_snv(data_file_cmt,odir)
if is_digit(str(exclude_samp)):
    samples_to_exclude=[]
    for s in dcs:
        if dcs[s]>=exclude_samp:
            samples_to_exclude.append(s)

else:
    if not exclude_samp=='':
        tem=re.split(',',exclude_samp)
        samples_to_exclude=tem
    else:
        samples_to_exclude=[""]
print('Exclude samples are:',samples_to_exclude)
#exit()

#samples_to_exclude=["AHM_v0002_D02","AHM_v0002_D03","AHM_v0012_F12"]


# Make subdirectory for this dataset
#dir_output = 'output_elife-sau'
dir_output=odir
os.system( "mkdir " + dir_output );


################ Run CNN first and then combine the result of CNN and default filters ######

####### Run CNN ########
cnn_pos,cnn_pred,cnn_prob,dgap=cnn.CNN_predict(data_file_cmt,data_file_cov,odir,samples_to_exclude,min_cov_samp) # The label is predicted by CNN
dlab=dict(zip(cnn_pos,cnn_pred)) # pos -> label
dprob=dict(zip(cnn_pos,cnn_prob)) # pos -> probability
#######  Done  #########
# Fast mode for cases where SNV positions > 1000000
if len(cnn_pos) > 100000:
    [quals,p,counts,in_outgroup,sample_names,indel_counter] = \
        snv.read_candidate_mutation_table_npz(data_file_cmt)
    if not len(in_outgroup)==len(sample_names):
        in_outgroup=np.array([False] * len(sample_names))
    my_cmt = snv.cmt_data_object( sample_names,
                             in_outgroup,
                             p,
                             counts,
                             quals,
                             indel_counter
                             )
    freq_d,check_d=snv.cal_freq_amb_samples(cnn_pos,my_cmt)
    if len(my_cmt.sample_names)>20:
        cutoff=0.1
    else:
        cutoff=0.25
    o=open(odir+'/pred_res.txt','w+')
    o.write('Pos\tPred\tProb')
    for name in my_cmt.sample_names:
        o.write('\t'+name)
    o.write('\n')
    pos_to_idx=dict(zip(my_cmt.p,range(len(my_cmt.p))))
    xc=0
    for c in cnn_pos:
        freq=freq_d[c]
        check=check_d[c]
        pred=str(cnn_pred[xc])
        prob=str(cnn_prob[xc])
        if pred=='0':
            if freq>cutoff or check:
                out_pred='0'
                out_prob=prob
            else:
                out_pred='1'
                out_prob=str(1-float(prob))
        else:
            out_pred=pred
            out_prob=prob
        idx=pos_to_idx[c]
        bases=snv.ints2nts(my_cmt.major_nt[:,idx])
        o.write(str(c)+'\t'+out_pred+'\t'+out_prob)
        for b in bases:
            o.write('\t'+b)
        o.write('\n')
        xc+=1
    print('Too many positions to predict (>100000), use a fast mode.......')

    samples_to_exclude_bool = np.array([x in samples_to_exclude for x in sample_names])
    keep_p = np.isin(my_cmt.p, cnn_pos)
    my_cmt_zero_rebuild = copy.deepcopy(my_cmt)
    my_cmt_zero_rebuild.filter_positions(keep_p)
    label = np.array([dlab[pos]==1 for pos in my_cmt_zero_rebuild.p])
    prob = np.array([dprob[pos] for pos in my_cmt_zero_rebuild.p])
    recomb = np.array([False] * len(my_cmt_zero_rebuild.p))
    quals_new = my_cmt_zero_rebuild.quals * -1
    new_cmt = {
        'sample_names': my_cmt_zero_rebuild.sample_names,
        'p': my_cmt_zero_rebuild.p,
        'counts': my_cmt_zero_rebuild.counts,
        'quals': quals_new,
        'in_outgroup': my_cmt_zero_rebuild.in_outgroup,
        'indel_counter': my_cmt_zero_rebuild.indel_stats,
        'prob': prob,
        'label': label,
        'recomb': recomb,
        'samples_exclude_bool': samples_to_exclude_bool,
    }
    np.savez_compressed(odir+'/candidate_mutation_table_final.npz', **new_cmt)
    exit()

#%% Generate candidate mutation table object

# Import candidate mutation table data generated in Snakemake

# Use this version for updated candidate mutation table matrices
[quals,p,counts,in_outgroup,sample_names,indel_counter] = \
    snv.read_candidate_mutation_table_npz(data_file_cmt) 
#dx=np.where(p==832924)
#print(counts[:,dx,:])
if not len(in_outgroup)==len(sample_names):
    in_outgroup=np.array([False] * len(sample_names))
#print(in_outgroup)
# # Use this version for old candidate mutation table matrices
# [quals,p,counts,in_outgroup,sample_names,indel_counter] = \
#     snv.read_old_candidate_mutation_table_pickle_gzip( data_file_cmt ) 


# Create instance of candidate mutation table class
my_cmt = snv.cmt_data_object( sample_names, 
                             in_outgroup, 
                             p, 
                             counts, 
                             quals, 
                             indel_counter 
                             )
#print(counts[:,0,:8])

#%% Import reference genome information

# Create instance of reference genome class
#print(dir_ref_genome)
#exit()
my_rg = snv.reference_genome_object( dir_ref_genome )
#exit()
#print(p.shape)
contig_p=my_rg.p2contigpos(p)
#print(contig_p.shape)
#print(my_rg.contig_names)
pre_a=re.split('_',my_rg.contig_names[0])
pre='_'.join(pre_a[:-1])
d={}
x=0
for c in contig_p:
	#print(c)
	d[p[x]]=[int(c[0]),int(c[1])]
	x+=1
#print(len(d))	
#print(d)
#exit()
'''
f=open(odir+'/cnn_res.txt','r')
o=open(odir+'/cnn_res_contig.txt','w+')
line=f.readline()
o.write('Contig\tContig_pos\tPos_info\tPredicted_label\tProbability\tGap_Filt\n')
while True:
	line=f.readline().strip()
	if not line:break
	ele=line.split('\t')
	o.write(str(my_rg.contig_names[d[int(ele[0])][0]-1])+'\t'+str(d[int(ele[0])][1])+'\t'+line+'\n')
o.close()
'''
#print(contig_p.shape,my_rg.contig_names.shape)
#exit()
my_rg_annot = my_rg.annotations
#print(my_rg_annot)
#my_rg_annot_0 = my_rg_annot[0]

#exit()

#%% Process raw coverage matrix

# Create instance of simple coverage class (just a summary of coverage matrix data to save memory)
# my_cov = snv.cov_data_object_simple( snv.read_cov_mat_npz( covFile ), 
#                                     my_cmt.sample_names, 
#                                     my_rg.genome_length, 
#                                     my_rg.contig_starts, 
#                                     my_rg.contig_names 
#                                     )

# Create instance of full coverage matrix class (includes full coverage matrix as attribute)
my_cov = snv.cov_data_object( snv.read_cov_mat_npz( data_file_cov ), \
                             my_cmt.sample_names, \
                             my_rg.genome_length, \
                             my_rg.contig_starts, \
                             my_rg.contig_names \
                             )



#%% Exclude any samples listed above as bad

samples_to_exclude_bool = np.array( [x in samples_to_exclude for x in my_cmt.sample_names] )
my_cmt_zero_rebuild=copy.deepcopy(my_cmt)
my_cmt.filter_samples( ~samples_to_exclude_bool )
my_cov.filter_samples( ~samples_to_exclude_bool )
my_cmt_zero=copy.deepcopy(my_cmt)
#my_cmt_for_rebuild=copy.deepcopy(my_cmt)
#print(my_cov)
#exit()
######################
## Dicts For Table  ##
######################
# dpt is used to keep the result of filters
dpt={} # dict used to keep information of identified SNPs. e.g d->pos -> {"cov_filter": 1,"qual_filter":1,...... }
#### We can't cause there are multiple samples for each position
# dft is used to keep the value of fwd reads
#dft={} # e.g. pos -> {"cov":3,"qual":30,...}
# drt is used to keep the value of rev reads
#drt={} # e.g. pos -> {"cov":3,"qual":30,...}


#%%###################
## FILTER BASECALLS ##
######################


#%% FILTER BASECALLS

# Create instance of basecalls class for initial calls
my_calls = snv.calls_object( my_cmt )
#my_calls_raw = snv.calls_object( my_cmt )
#my_calls_raw_zero=copy.deepcopy(my_calls)
'''
x=my_calls_raw_zero.get_frac_Ns_by_position()
print(my_calls.p[-10:])
print(my_cmt.major_nt[:,-10:])
print(my_cmt.major_nt_freq[:,-10:])
#print(my_cmt.major_nt_freq.shape)
print(my_cmt.minor_nt[:,-10:])
print(my_cmt.minor_nt_freq[:,-10:])
#print(my_cmt.minor_nt_freq.shape)
#print(x.shape,x,my_calls.p)
exit()
'''


#%% Filter parameters

# Remove samples that are not high quality
filter_parameter_sample_across_sites = {
                                        'min_average_coverage_to_include_sample': 0, # remove samples that have low coverage # default: 10
                                        'max_frac_Ns_to_include_sample': 1 # remove samples that have too many undefined base (ie. N). # default: 0.2
                                        }

# Remove sites within samples that are not high quality
filter_parameter_site_per_sample = {
                                    'min_major_nt_freq_for_call' : 0.85, # on individual samples, a calls' major allele must have at least this freq
                                    'min_cov_per_strand_for_call' : min_cov_filt,  # on individual samples, calls must have at least this many reads on the fwd/rev strands individually
                                    'min_qual_for_call' : 30, # on individual samples, calls must have this minimum quality score
                                    'max_frac_reads_supporting_indel' : 0.33 # on individual samples, no more than this fraction of reads can support an indel at any given position
                                    }

# Remove sites across samples that are not high quality
filter_parameter_site_across_samples = {
                                        'max_fraction_ambigious_samples' : 1, # across samples per position, the fraction of samples that can have undefined bases
                                        'min_median_coverage_position' : 5, # across samples per position, the median coverage
                                        'max_mean_copynum' : 4, # mean copy number at a positions across all samples
                                        'max_max_copynum' : 7 # max maximum copynumber that a site can have across all samples
                                        }

original_stdout = sys.stdout
sys.stdout = Tee(odir+'/pipe_log.txt')
#%% Filter samples based on coverage

# Identify samples with low coverage and make a histogram
[ low_cov_samples, goodsamples_coverage ] = snv.filter_samples_by_coverage( 
    my_cov.get_median_cov_of_chromosome(), 
    filter_parameter_sample_across_sites['min_average_coverage_to_include_sample'], 
    my_cov.sample_names, 
    True, 
    dir_output 
    )
#print(low_cov_samples)
#exit()

# Filter candidate mutation table, coverage, and calls objects
my_cmt.filter_samples( goodsamples_coverage )
my_cov.filter_samples( goodsamples_coverage )
my_calls.filter_samples( goodsamples_coverage )


#%% Filter calls per position per sample
#print(my_cmt.p)
#print(my_calls.calls.T)
#print(my_cmt.fwd_cov.T)
#print(my_cmt.rev_cov.T)
#print(my_cmt)

#print('---------')
#print(my_calls.calls.T.shape)

# # Filter based on quality
# my_calls_raw=copy.deepcopy(my_calls)
# my_calls_qual=copy.deepcopy(my_calls)
# #my_cmt_qual=copy.deepcopy(my_cmt)
# my_calls_qual.filter_calls_by_element(
#     my_cmt.quals < filter_parameter_site_per_sample['min_qual_for_call']
#     ) # quality too low
#
# tokens = snv.token_generate(my_calls_raw.calls.T, my_calls_qual.calls.T, 'filter-qual')
# dpt['qual']=dict(zip(my_calls_qual.p,tokens))


# Filter based on quality
my_calls_raw=copy.deepcopy(my_calls)
my_calls.filter_calls_by_element(
    my_cmt.quals < filter_parameter_site_per_sample['min_qual_for_call']
    ) # quality too low
tokens = snv.token_generate(my_calls_raw.calls.T, my_calls.calls.T, 'filter-qual')
dpt['qual']=dict(zip(my_calls.p,tokens))

# Filter based on coverage
my_calls_raw=copy.deepcopy(my_calls)

my_calls.filter_calls_by_element( 
    my_cmt.fwd_cov < filter_parameter_site_per_sample['min_cov_per_strand_for_call'] 
    ) # forward strand coverage too low

my_calls.filter_calls_by_element( 
    my_cmt.rev_cov < filter_parameter_site_per_sample['min_cov_per_strand_for_call'] 
    ) # reverse strand coverage too low
tokens = snv.token_generate(my_calls_raw.calls.T, my_calls.calls.T, 'filter-coverage')
dpt['cov']=dict(zip(my_calls.p,tokens))
#print(my_cmt.quals.shape)
#exit()
#print(my_cmt.p,len(tokens))
#print(dpt)
#exit()
#print(my_calls.calls.T)
#print(my_calls_raw.calls.T)
#print(np.array_equal(my_calls.calls,my_calls_raw.calls) )
#unequal_mask = my_calls.calls != my_calls_raw.calls
#print(unequal_mask)
#exit()

#print('---------')

#tokens = snv.token_generate(my_calls_raw.calls.T, my_calls.calls.T, 'filter-qual')
#dpt['qual']=dict(zip(my_calls.p,tokens))

#exit()
#print(my_calls.p)
#print(my_calls.calls.T)
#print(my_calls.calls.T.shape)
#exit()
# Filter based on major allele frequency
my_calls_raw=copy.deepcopy(my_calls)
my_calls.filter_calls_by_element( 
    my_cmt.major_nt_freq < filter_parameter_site_per_sample['min_major_nt_freq_for_call'] 
    ) # major allele frequency too low
tokens = snv.token_generate(my_calls_raw.calls.T, my_calls.calls.T, 'filter-major allele freq')
dpt['maf']=dict(zip(my_calls.p,tokens))

#%% Filter positions with indels

with np.errstate(divide='ignore',invalid='ignore'):
    # compute number of reads supporting an indel
    frac_reads_supporting_indel = np.sum(my_cmt.indel_stats,axis=2)/my_cmt.coverage # sum reads supporting insertion plus reads supporting deletion
    frac_reads_supporting_indel[ ~np.isfinite(frac_reads_supporting_indel) ] = 0
    # note: this fraction can be above zero beacuse the number of reads supporting an indel includes a +/-3 bp window around a given position on the genome
my_calls_raw=copy.deepcopy(my_calls)
my_calls.filter_calls_by_element( 
    frac_reads_supporting_indel > filter_parameter_site_per_sample['max_frac_reads_supporting_indel'] 
    ) # too many reads supporting indels
tokens = snv.token_generate(my_calls_raw.calls.T, my_calls.calls.T, 'filter-indel')
dpt['indel']=dict(zip(my_calls.p,tokens))

#%% Filter positions that look iffy across samples
my_calls_raw=copy.deepcopy(my_calls)
my_calls.filter_calls_by_position( 
    my_calls.get_frac_Ns_by_position() > filter_parameter_site_across_samples['max_fraction_ambigious_samples'] 
    ) # too many samples with ambiuguous calls at this position
tokens = snv.token_generate(my_calls_raw.calls.T, my_calls.calls.T, 'filter-max_fraction_ambigious_samples')
dpt['mfas']=dict(zip(my_calls.p,tokens))

snv.filter_histogram( 
    my_calls.get_frac_Ns_by_position(), 
    filter_parameter_site_across_samples['max_fraction_ambigious_samples'], 
    'Fraction Ns by position'
    )

my_calls_raw=copy.deepcopy(my_calls)
my_calls.filter_calls_by_position( 
    np.median( my_cmt.coverage, axis=0 ) < filter_parameter_site_across_samples['min_median_coverage_position'] 
    ) # insufficient median coverage across samples at this position
tokens = snv.token_generate(my_calls_raw.calls.T, my_calls.calls.T, 'filter-min_median_coverage_position')
dpt['mmcp']=dict(zip(my_calls.p,tokens))
# # Optional code to make a histogram
# snv.filter_histogram( 
#     np.median( my_cmt.coverage, axis=0 ), 
#     filter_parameter_site_across_samples['min_median_coverage_position'], 
#     'Median coverage by position'
#     )
my_calls_raw=copy.deepcopy(my_calls)
copy_number_per_sample_per_pos = my_cmt.coverage / np.expand_dims( my_cov.get_median_cov_of_chromosome(), 1) # compute copy number
copy_number_avg_per_pos = np.mean( copy_number_per_sample_per_pos, axis=0 ) # mean copy number at each position
copy_number_avg_per_pos[np.isnan(copy_number_avg_per_pos)]=0
my_calls.filter_calls_by_position( 
    copy_number_avg_per_pos > filter_parameter_site_across_samples['max_mean_copynum'] 
    ) # average copy number too high
copy_number_max_per_pos = np.max( copy_number_per_sample_per_pos, axis=0 ) # mean copy number at each position
copy_number_max_per_pos[np.isnan(copy_number_max_per_pos)]=0
my_calls.filter_calls_by_position( 
    copy_number_max_per_pos > filter_parameter_site_across_samples['max_max_copynum'] 
    ) # average copy number too high
tokens = snv.token_generate(my_calls_raw.calls.T, my_calls.calls.T, 'filter-copy number')
dpt['cpn']=dict(zip(my_calls.p,tokens))
#print(my_calls.p.shape)
#print(my_calls.p)
#exit()
# # Optional code to make a histogram
# snv.filter_histogram( 
#     copy_number_avg_per_pos, 
#     filter_parameter_site_across_samples['max_mean_copynum'] , 
#     'Average copy number by position'
#     )

#%% Filter samples that have too many ambiguous calls

# Identify samples with many ambiguous basecalls and make a histogram

pos_to_consider = my_calls.p[ np.any(  my_calls.calls, axis=0 ) ] # mask positions with no basecalls in any samples
[ samples_with_toomanyNs, goodsamples_nonambig ] = snv.filter_samples_by_ambiguous_basecalls( 
    my_calls.get_frac_Ns_by_sample( pos_to_consider ), 
    filter_parameter_sample_across_sites['max_frac_Ns_to_include_sample'], 
    my_calls.sample_names, 
    my_calls.in_outgroup, # does not filter outgroup samples!!!
    True, 
    dir_output 
    )
#print(len(pos_to_consider))
#print(my_calls.get_frac_Ns_by_sample( pos_to_consider ))
#print(my_calls.sample_names)
#print(samples_with_toomanyNs)
#exit()

#print(my_cmt.p.shape)
# Filter candidate mutation table, coverage, and calls objects
#my_cmt.filter_samples( goodsamples_nonambig )
#my_cov.filter_samples( goodsamples_nonambig )
#my_calls.filter_samples( goodsamples_nonambig )
#print(my_cmt.p.shape)
#exit()


#%%##########################
## INFER ANCESTRAL ALLELES ##
#############################


#%% Part 1: Get ancestral nucleotide from outgroup

# Ancestral alleles should be inferred from an outgroup.

# Filtered calls for outgroup samples only
calls_outgroup = my_calls.get_calls_in_outgroup_only()
# Switch N's (0's) to NaNs
calls_outgroup_N_as_NaN = calls_outgroup.astype('float') # init ()
calls_outgroup_N_as_NaN[ calls_outgroup_N_as_NaN==0 ] = np.nan

# Infer ancestral allele as the most common allele among outgroup samples (could be N)
calls_ancestral = np.zeros( my_calls.num_pos, dtype='int') # init as N's
outgroup_pos_with_calls = np.any(calls_outgroup,axis=0) # positions where the outgroup has calls
calls_ancestral[outgroup_pos_with_calls] = stats.mode( calls_outgroup_N_as_NaN[:,outgroup_pos_with_calls], axis=0, nan_policy='omit' ).mode.squeeze()

# Report number of ancestral alleles inferred from outgroup
print('Number of candidate SNVs with outgroup alleles: ' + str(sum(outgroup_pos_with_calls)) + '.')
print('Number of candidate SNVs missing outgroup alleles: ' + str(sum(calls_ancestral==0)) + '.')


#%% Part 2: Fill in any missing data with nucleotide from reference

# WARNING! Rely on this method with caution especially if the reference genome 
# was derived from one of your ingroup samples.

# # Pull alleles from reference genome across p
calls_reference = my_rg.get_ref_NTs_as_ints( p )

# # Update ancestral alleles
pos_to_update = ( calls_ancestral==0 )
calls_ancestral[ pos_to_update ] = calls_reference[ pos_to_update ]



#%%#########################
## IDENTIFY SNV POSITIONS ##
############################


#%% Compute mutation quality

# Grab filtered calls from ingroup samples only
calls_ingroup = my_calls.get_calls_in_sample_subset( np.logical_not( my_calls.in_outgroup ) )
quals_ingroup = my_cmt.quals[ np.logical_not( my_calls.in_outgroup ),: ]
num_samples_ingroup = sum( np.logical_not( my_calls.in_outgroup ) )
# Note: Here we are only looking for SNV differences among ingroup samples. If
# you also want to find SNV differences between the ingroup and the outgroup
# samples (eg mutations that have fixed across the ingroup), then you need to
# use calls and quals matrices that include outgroup samples.

# Compute quality
[ mut_qual, mut_qual_samples ] = snv.compute_mutation_quality( calls_ingroup, quals_ingroup ) 
# note: returns NaN if there is only one type of non-N call


#%% Identify suspected recombination positions

# Filter params
filter_parameter_recombination = {
                                    'distance_for_nonsnp' : 1000, # region in bp on either side of goodpos that is considered for recombination
                                    'corr_threshold_recombination' : 0.75 # minimum threshold for correlation
                                    }

# Find SNVs that are are likely in recombinant regions
[ p_recombo, recombo_bool ] = snv.find_recombination_positions( \
    my_calls, my_cmt, calls_ancestral, mut_qual, my_rg, \
    filter_parameter_recombination['distance_for_nonsnp'], \
    filter_parameter_recombination['corr_threshold_recombination'], \
    True, dir_output \
    )

#print(recombo_bool.shape,my_calls.p.shape)
dpt['recomb']=dict(zip(my_calls.p,recombo_bool))
#exit()
    
# Save positions with likely recombination
if len(p_recombo)>0:
    with open( dir_output + '/snvs_from_recombo.csv', 'w') as f:
        for p in p_recombo:
            f.write(str(p)+'\n')



#%% Determine which positions have high-quality SNVs

# Filters
filter_SNVs_not_N = ( calls_ingroup != snv.nts2ints('N') ) # mutations must have a basecall (not N)
filter_SNVs_not_ancestral_allele = ( calls_ingroup != np.tile( calls_ancestral, (num_samples_ingroup,1) ) ) # mutations must differ from the ancestral allele
filter_SNVs_quals_not_NaN = ( np.tile( mut_qual, (num_samples_ingroup,1) ) >= 1) # alleles must have strong support 
filter_SNVs_not_recombo = np.tile( np.logical_not(recombo_bool), (num_samples_ingroup,1) ) # mutations must not be due to suspected recombination

# Fixed mutations per sample per position
fixedmutation = \
    filter_SNVs_not_N \
    & filter_SNVs_not_ancestral_allele \
    & filter_SNVs_quals_not_NaN \
    & filter_SNVs_not_recombo # boolean

#print(fixedmutation.shape)
#exit()

hasmutation = np.any( fixedmutation, axis=0) # boolean over positions (true if at lest one sample has a mutation at this position)

goodpos_bool = np.any( fixedmutation, axis=0 )
#print(goodpos_bool.shape)
#exit()
goodpos_idx = np.where( goodpos_bool )[0]
#print(goodpos_idx)
tokens_final=snv.generate_tokens_last(tokens,goodpos_idx,'filter-fixedmutation')
dpt['fix']=dict(zip(my_calls.p,tokens_final))
#exit()
num_goodpos = len(goodpos_idx)
print('Num mutations identified by WideVariant: '+str(num_goodpos))

####### Combine CNN and WideVariant output and generate the SNV information table #######
goodpos_idx_cnn=cnn_pos[np.where(cnn_pred==1)]
#print(goodpos_idx_cnn)
#exit()
goodpos_idx_wd=my_calls.p[goodpos_idx]
all_p=np.sort(np.union1d(goodpos_idx_cnn,goodpos_idx_wd))
#all_p=my_calls.p[all_p]
#print(all_p)
#print(np.where(my_cmt.p==1095218))
#exit()
goodpos_bool,goodpos_bool_all=snv.generate_cnn_filter_table(all_p,goodpos_idx_wd,dpt,dlab,dprob,dir_output,my_cmt.p,dgap,my_cmt_zero,min_cov_filt)
goodpos_idx = np.where( goodpos_bool )[0]
#print(goodpos_bool.shape)
#exit()
#print(goodpos_bool,goodpos_bool.shape)
#print(goodpos_idx)
#exit()
goodpos_idx_all = np.where( goodpos_bool_all)[0]
num_goodpos = len(goodpos_idx)
num_goodpos_all = len(goodpos_idx_all)
print('Num mutations identified by CNN+WideVariant+Recomb_filt: '+str(num_goodpos_all))
print('Num mutations identified by CNN+Recomb_filt:'+str(num_goodpos))
if greport==0:
    print('Find -g 0. No report and extra files will be generated...')
    exit()
#pos_to_consider = my_calls.p[ np.any(  my_calls.calls, axis=0 ) ] # mask positions with no basecalls in any samples

[ samples_with_toomanyNs, goodsamples_nonambig ] = snv.filter_samples_by_ambiguous_basecalls(
    my_calls.get_frac_Ns_by_sample( goodpos_idx ),
    filter_parameter_sample_across_sites['max_frac_Ns_to_include_sample'],
    my_calls.sample_names,
    my_calls.in_outgroup, # does not filter outgroup samples!!!
    True,
    dir_output
    )
#print(my_calls.sample_names)
#print(my_calls.get_frac_Ns_by_sample( goodpos_idx ))
#print(samples_with_toomanyNs)
#exit()


#%% Make and annotate a SNV table

# Prepare data for SNV table

my_cmt_goodpos = my_cmt.copy()
my_cmt_goodpos.filter_positions( goodpos_bool )
my_cmt_goodpos_ingroup = my_cmt_goodpos.copy()
my_cmt_goodpos_ingroup.filter_samples( np.logical_not( my_cmt_goodpos_ingroup.in_outgroup ) )

my_calls_goodpos = my_calls.copy()
my_calls_goodpos.filter_positions( goodpos_bool )
calls_goodpos = my_calls_goodpos.calls
calls_goodpos_ingroup = calls_goodpos[ np.logical_not( my_calls_goodpos.in_outgroup ),: ]

p_goodpos = my_calls_goodpos.p

calls_ancestral_goodpos = calls_ancestral[ goodpos_bool ]

# Only for bar charts
my_cmt_goodpos_all = my_cmt.copy()
#print(np.where(my_cmt.p==1095218))
#print(goodpos_bool_all[np.where(my_cmt.p==1095218)[0]])
my_cmt_goodpos_all.filter_positions( goodpos_bool_all )
#print(my_cmt_goodpos_all.p)
#exit()
my_cmt_goodpos_ingroup_all = my_cmt_goodpos_all.copy()
my_cmt_goodpos_ingroup_all.filter_samples( np.logical_not( my_cmt_goodpos_ingroup_all.in_outgroup ) )

my_calls_goodpos_all= my_calls.copy()
my_calls_goodpos_all.filter_positions( goodpos_bool_all )
calls_goodpos_all = my_calls_goodpos_all.calls
calls_goodpos_ingroup_all = calls_goodpos_all[ np.logical_not( my_calls_goodpos_all.in_outgroup ),: ]
p_goodpos_all = my_calls_goodpos_all.p
calls_ancestral_goodpos_all = calls_ancestral[ goodpos_bool_all ]

# Generate SNV table

# Parameters
promotersize = 250; # how far upstream of the nearest gene to annotate something a promoter mutation (not used if no annotation)

# Make a table (pandas dataframe) of SNV positions and relevant annotations
mutations_annotated = snv.annotate_mutations( \
    my_rg, \
    p_goodpos_all, \
    np.tile( calls_ancestral[goodpos_idx_all], (my_cmt_goodpos_ingroup_all.num_samples,1) ), \
    calls_goodpos_ingroup_all, \
    my_cmt_goodpos_ingroup_all, \
    fixedmutation[:,goodpos_idx_all], \
    mut_qual[:,goodpos_bool_all].flatten(), \
    promotersize \
    ) 

#print(mutations_annotated)
#exit()
#%% SNV quality control plots

# Note: These data visualizations are intended to help you evaluate if your SNV
# filters are appropriate. Do not proceed to the next step until you are 
# convinced that your filters are strict enough to filter low-quality SNVs, but
# not so strict that good SNVs are eliminated as well. 


#print(p_goodpos_all)
#exit()
# Clickable bar charts for each SNV position

# If pos>2000, then only first 2000 charts will be plotted
chart_limit = 2000
if num_goodpos_all > chart_limit:
    idx_slice = slice(0, chart_limit)
else:
    idx_slice = slice(None)
snv.plot_interactive_scatter_barplots(
    p_goodpos_all[idx_slice],
    mut_qual[0, goodpos_idx_all][idx_slice],
    'pos',
    'qual',
    my_cmt_goodpos_all.sample_names,
    mutations_annotated.iloc[idx_slice],
    my_cmt_goodpos_all.counts[:, idx_slice, :],
    dir_output,
    False)

# Old bar char plotting function
#snv.plot_interactive_scatter_barplots( \
#    p_goodpos_all, \
#    mut_qual[0,goodpos_idx_all], \
#    'pos', \
#    'qual', \
#    my_cmt_goodpos_all.sample_names, \
#    mutations_annotated, \
#    my_cmt_goodpos_all.counts,dir_output,False)

#exit()
# Heatmaps of basecalls, coverage, and quality over SNV positions
if 300>num_goodpos_all>0:
    snv.make_calls_qc_heatmaps( my_cmt_goodpos_all, my_calls_goodpos_all, True, dir_output,False )



#%%###########################
## PARSIMONY AND TREEMAKING ##
##############################

#%% Filter calls for tree

# Note: Here we are using looser filters than before

# Choose subset of samples or positions to use in the tree by idx
samplestoplot = np.arange(my_cmt_goodpos_all.num_samples) # default is to use all samples 
goodpos4tree = np.arange(num_goodpos_all) # default is to use all positions
#print(goo)

# Get calls for tree
my_calls_tree = snv.calls_object( my_cmt_goodpos_all ) # re-initialize calls

# Apply looser filters than before (want as many alleles as possible)
filter_parameter_calls_for_tree = {
                                    'min_cov_for_call' : 1, # on individual samples, calls must have at least this many fwd+rev reads
                                    'min_qual_for_call' : 30, # on individual samples, calls must have this minimum quality score
                                    'min_major_nt_freq_for_call' : 0.75,  # on individual samples, a call's major allele must have at least this freq
                                    }

my_calls_tree.filter_calls_by_element( 
    my_cmt_goodpos_all.coverage < filter_parameter_calls_for_tree['min_cov_for_call'] 
    ) # forward strand coverage too low

my_calls_tree.filter_calls_by_element( 
    my_cmt_goodpos_all.quals < filter_parameter_calls_for_tree['min_qual_for_call'] 
    ) # quality too low

my_calls_tree.filter_calls_by_element( 
    my_cmt_goodpos_all.major_nt_freq < filter_parameter_calls_for_tree['min_major_nt_freq_for_call'] 
    ) # major allele frequency too low

# Make QC plots again using calls for the tree
if 300>num_goodpos_all>0:
    snv.make_calls_qc_heatmaps( my_cmt_goodpos_all, my_calls_tree, False, dir_output, False )

# %% Make a tree

if num_goodpos_all > 0:

    # Get nucleotides of basecalls (ints to NTs)
    calls_for_treei = my_calls_tree.calls[
        np.ix_(samplestoplot, goodpos4tree)]  # numpy broadcasting of row_array and col_array requires np.ix_()
    calls_for_tree = snv.ints2nts(calls_for_treei)  # NATCG translation

    # Sample names for tree
    treesampleNamesLong = my_cmt_goodpos_all.sample_names
    for i, samplename in enumerate(treesampleNamesLong):
        if not samplename[0].isalpha():
            treesampleNamesLong[i] = 'S' + treesampleNamesLong[
                i]  # sample names are modified to make parsing easier downstream
    sampleNamesDnapars = ["{:010d}".format(i) for i in range(my_cmt_goodpos_all.num_samples)]

    # Add inferred ancestor and reference
    calls_ancestral_for_tree = np.expand_dims(snv.ints2nts(calls_ancestral_goodpos_all), axis=0)
    calls_reference_for_tree = np.expand_dims(my_rg.get_ref_NTs(my_calls_tree.p), axis=0)
    calls_for_tree_all = np.concatenate((calls_ancestral_for_tree, calls_reference_for_tree, calls_for_tree),
                                        axis=0)  # first column now outgroup_nts; outgroup_nts[:, None] to make ndims (2) same for both
    sampleNamesDnapars_all = np.append(['Sanc', 'Sref'], sampleNamesDnapars)
    treesampleNamesLong_all = np.append(['inferred_ancestor', 'reference_genome'], treesampleNamesLong)
    
    try:
        # Build tree
        snv.generate_tree( \
            calls_for_tree_all.transpose(), \
            treesampleNamesLong_all, \
            sampleNamesDnapars_all, \
            ref_genome_name, \
            dir_output, \
            "snv_tree_" + ref_genome_name, \
            buildTree='PS' \
            )
    except Exception as e:
        print('#### error skip #####: something wrong in snv.generate_tree... skip...')
        print(f"Error message: {str(e)}")
        traceback.print_exc()



#%%###################
## SAVE DATA TABLES ##
######################


#%% Write SNV table to a tsv file

# Note: important to use calls for tree (rather than calls for finding SNVs)

# Make table
if num_goodpos>0:
    # This is the raw mutation table - contain positions identified by both CNN and WD, and are not recombinations
    output_tsv_filename = dir_output + '/' + 'snv_table_mutations_annotations_raw.tsv'
    snv.write_mutation_table_as_tsv( \
        p_goodpos_all, \
        mut_qual[0,goodpos_idx_all], \
        my_cmt_goodpos_all.sample_names, \
        mutations_annotated, \
        calls_for_tree, \
        treesampleNamesLong, \
        output_tsv_filename \
        
        )
    out_merge_tsv=dir_output+'/snv_table_merge_all_mut_annotations_draft.tsv'
    snv.merge_two_tables(dir_output+'/snv_table_cnn_plus_filter.txt',output_tsv_filename,out_merge_tsv)
    #exit()
    snv.generate_html_with_thumbnails(dir_output+'/snv_table_merge_all_mut_annotations_draft.tsv', dir_output+'/snv_table_with_charts_draft.html', dir_output+'/bar_charts')
    # Generate the tree for each identified SNPs
    try:
        bst.mutationtypes(dir_output+"/snv_tree_genome_latest.nwk.tree",dir_output+'/snv_table_merge_all_mut_annotations_draft.tsv',1,dir_output)
    except Exception as e:
        print('#### error skip #####: something wrong in bst.mutationtypes... skip...')
        print(f"Error message: {str(e)}")
        traceback.print_exc()
    # # Contain all positions identified by CNN or WideVariant - even those false positions
    # snv.write_mutation_table_as_tsv( \
    #     p_goodpos_all, \
    #     mut_qual[0, goodpos_idx_all], \
    #     my_cmt_goodpos_all.sample_names, \
    #     mutations_annotated, \
    #     calls_for_tree, \
    #     treesampleNamesLong, \
    #     output_tsv_filename \
    #     )

# Rebuild output for new requirement - 2025-03-21 - Herui
'''
update - 1 - save candidate mutation table with only label '1' pos
update - 2 - regenerate output text and html report
update - 3 - add dN/dS output
'''
f=open(dir_output+'/snv_table_merge_all_mut_annotations_draft.tsv','r')
o=open(dir_output+'/snv_table_merge_all_mut_annotations_final.tsv','w+')
o2=open(dir_output+'/snv_table_merge_all_mut_annotations_label0.tsv','w+')
line=f.readline()
o.write(line)
o2.write(line)
dk={}
dl={}
dr={}
while True:
    line=f.readline().strip()
    if not line:break
    ele=line.split('\t')
    if int(ele[1])==0:
        #continue
        o2.write(line+'\n')
        dl[int(ele[0])]=''
        if ele[4]=='skip':
            dk[int(ele[0])]=0
        else:
            dk[int(ele[0])]=float(ele[4])
    else:
        o.write(line+'\n')
        if ele[4]=='skip':
            dk[int(ele[0])]=0
        else:
            dk[int(ele[0])]=float(ele[4])
    if int(ele[13])==1:
        dr[int(ele[0])]=int(ele[13])
o.close()
o2.close()
snv.generate_html_with_thumbnails(dir_output+'/snv_table_merge_all_mut_annotations_final.tsv', dir_output+'/snv_table_with_charts_final.html', dir_output+'/bar_charts')
if len(dl)>0:
    snv.generate_html_with_thumbnails(dir_output+'/snv_table_merge_all_mut_annotations_label0.tsv', dir_output+'/snv_table_with_charts_label0.html', dir_output+'/bar_charts')
keep_p=[]
prob=[]
label=[]
recomb=[]
for s in my_cmt_zero_rebuild.p:
    '''
    if s in dl:
        label.append(False)
    else:
        label.append(True)
    '''
    if s in dk:
        keep_p.append(True)
        prob.append(dk[s])
        if s in dl:
            label.append(False)
        else:
            label.append(True)
        if s in dr:
            recomb.append(True)
        else:
            recomb.append(False)
    else:
        keep_p.append(False)
    '''
    if s in dr:
        recomb.append(True)
    else:
        recomb.append(False)
    '''


keep_p=np.array(keep_p)
my_cmt_zero_rebuild.filter_positions(keep_p)
label=np.array(label)
recomb=np.array(recomb)
quals_new=my_cmt_zero_rebuild.quals* -1
new_cmt={'sample_names': my_cmt_zero_rebuild.sample_names,'p':my_cmt_zero_rebuild.p,'counts':my_cmt_zero_rebuild.counts,'quals':quals_new,'in_outgroup':my_cmt_zero_rebuild.in_outgroup,'indel_counter':my_cmt_zero_rebuild.indel_stats,'prob':prob,'label':label,'recomb':recomb,'samples_exclude_bool':samples_to_exclude_bool}
np.savez_compressed(dir_output+'/candidate_mutation_table_final.npz', **new_cmt)

# dN/dS part
sys.path.insert(0, './miniscripts_for_dNdS')
dir_ref_genome = refg
annotation_full = pd.read_csv(dir_output+'/snv_table_merge_all_mut_annotations_final.tsv',sep='\t')
output_directory = dir_output+'/dNdS_out'
if not os.path.exists(output_directory):
    os.makedirs(output_directory)
# Mutation spectrum
NTs = np.array(['A', 'T', 'C', 'G', 'N'])
mutationmatrix, mut_observed, typecounts, prob_nonsyn = snv.mutation_spectrum_module(annotation_full, NTs)
# Notes:
# * Mutation spectrum tallied for each mutant alelle
# * Mutation type only tallied when there is only one mutant allele
# * Assumes all mutants are true de novo mutants
num_muts_for_empirical_spect = 100
if np.sum(mut_observed) >= num_muts_for_empirical_spect:
    # attempt to get mutation spectrum empirically if there were enough mutations
    mut_spec_prob = mut_observed / np.sum(mut_observed)
else:
    # otherwise assume a uniform mutation spectrum
    mut_spec_prob = np.ones(mut_observed.size) / mut_observed.size
    print('Warning! Assuming uniform mutation spectrum.')

# Expected N/S
probnonsyn_expected = snv.compute_expected_dnds(dir_ref_genome, mut_spec_prob)

# this is going to be a bit off from matlab bc of how it deals with alternate start codons
dnds_expected = probnonsyn_expected / (1 - probnonsyn_expected)  # N/S expected from neutral model

# compute observed N/S for fixed and diverse mutations
    # define gene_nums_of_interest if binning mutations for dN/dS analysis, rather than whole-genome dN/dS
p_nonsyn, CI_nonsyn, num_muts_N, num_muts_S = snv.compute_observed_dnds(annotation_full, gene_nums_of_interest=None)
dnds_observed = p_nonsyn / (1 - p_nonsyn)  # N/S observed
# note that in matlab version, binom(0,0) gives CI of [0,1] even when p=NaN. In python, both are NaN

# dN/dS
# relative to neutral model for this KEGG category
dNdS = dnds_observed / dnds_expected

CI_lower = (CI_nonsyn[0] / (1 - CI_nonsyn[0])) / dnds_expected
try:
    CI_upper = (CI_nonsyn[1] / (1 - CI_nonsyn[1])) / dnds_expected
except ZeroDivisionError:
    CI_upper = np.inf


print('dN/dS =', dNdS)

# save output as binary file using pickle
output_dict = {
    'dNdS': dNdS,
    'CI_lower': CI_lower,
    'CI_upper': CI_upper,
    'num_muts_N': num_muts_N,
    'num_muts_S': num_muts_S,
    'p_nonsyn': p_nonsyn,
    'probnonsyn_expected': probnonsyn_expected
}
'''
with open(output_directory+'/data_dNdS.pickle', 'wb') as f:
    pickle.dump(output_dict, f)
'''
np.savez_compressed(output_directory+'/data_dNdS.npz', **output_dict)

exit()



#%% Write table of tree distances to a csv file

if num_goodpos>0:
    
    # Compute number of SNVs separating each sample from the inferred ancestor
    fixedmutation_tree = ( calls_for_tree != calls_ancestral_for_tree ) & ( calls_for_tree != 'N' ) # boolean
    
    dist_to_anc_by_sample = np.sum( fixedmutation_tree, axis=1 )
    # Save to a file
    with open( dir_output + '/snv_table_tree_distances.csv', 'w') as f:
        f.write('sample_name,num_SNVs_to_ancestor\n')
        for i,name in enumerate(treesampleNamesLong):
            f.write(name + ',' + str(dist_to_anc_by_sample[i]) + '\n' )


#%% Write a table of summary stats

print( 'Number of samples in ingroup: ' + str(num_samples_ingroup) + '.') 
print( 'Number of good SNVs found: ' + str(num_goodpos) + '.') 
print('Number of good SNVs with outgroup alleles: ' + str(sum(calls_ancestral_goodpos>0)) + '.')
print('Number of good SNVs missing outgroup alleles: ' + str(sum(calls_ancestral_goodpos==0)) + '.')
if num_goodpos>0:
    dist_to_anc_by_sample_ingroup = dist_to_anc_by_sample[ np.logical_not(my_calls_tree.in_outgroup) ]
    dmrca_median = np.median( dist_to_anc_by_sample_ingroup )
    dmrca_min = np.min( dist_to_anc_by_sample_ingroup )
    dmrca_max = np.max( dist_to_anc_by_sample_ingroup )
else:
    dmrca_median = 0
    dmrca_min = 0
    dmrca_max = 0
print( 'Median dMRCA: ' + str(dmrca_median) + '.') 
print( 'Min dMRCA: ' + str(dmrca_min) + '.') 
print( 'Max dMRCA: ' + str(dmrca_max) + '.') 

with open( dir_output + '/snv_table_simple_stats.csv', 'w') as f:
    f.write('dataset,genome,num_samples_ingroup,num_snvs,num_snvs_with_outgroup_allele,dmrca_median,dmrca_min,dmrca_max,\n')
    f.write(dataset_name+','+ref_genome_name+','+str(num_samples_ingroup)+','+str(num_goodpos)+','+str(sum(calls_ancestral_goodpos>0))+','+str(dmrca_median)+','+str(dmrca_min)+','+str(dmrca_max)+',\n')



#%%##################
## CONTIG ANALYSIS ##
#####################

# This section examines contig coverage.


#%% Make a plot of presence/absence of each contig

my_cov.plot_heatmap_contig_copy_num( dir_output ,False)

# Record data in files
snv.write_generic_csv( my_cov.median_coverage_by_contig, my_cov.contig_names, my_cov.sample_names, dir_output+'/contig_table_median_coverage.csv' )
snv.write_generic_csv( my_cov.copy_number_by_contig, my_cov.contig_names, my_cov.sample_names, dir_output+'/contig_table_copy_number.csv' )


#%% Make plots of coverage traces of each contig

if type(my_cov) == snv.cov_data_object:
    for contig_num in np.linspace(1,my_cov.num_contigs,my_cov.num_contigs).astype(int):
        print('Generating copy number traces for contig ' + str(contig_num) + '...')
        my_cov.make_coverage_trace(contig_num,100,dir_output,False);
elif type(my_cov) == snv.cov_data_object_simple:
    print('Coverage matrix object type is cov_data_obj_simple, not cov_data_obj. Raw coverage matrix not available for copy number traces.')



'''
Inputs:
    paths_to_samples

Outputs:
    alignment_stats.csv

Change Log:
    2022.10.13, Arolyn: changed hardcoded directory structure to be compatible with GUS (though ideally this should not be hardcoded in the first place)
'''
import math
import os
import shutil
import argparse
import pandas as pd
import numpy as np
import matplotlib; matplotlib.use('agg')
import matplotlib.pyplot as plt

parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
                                description='''\
                            Generates QC stats on the bowtie2 rule
                               ''',
                               epilog="Questions or comments? --> aholton@mit.edu")
parser.add_argument("-s", dest="samples", help="csv file with samples",required=True,action='store')
parser.add_argument("-r", dest="refGenome", help="name of reference genome as string",required=True,action='store')
parser.add_argument("-d", dest="currentDirectory", help="current directory of Snakefile", required=True, action="store")
parser.add_argument("-o", dest="outfileString", help="name of outfile without file extension", required=True, action="store")
args = parser.parse_args()

def main(path_to_samples, reference_genome, current_directory, out_file_string):

    cwd = current_directory
    home_dir = cwd.split("/")[-1]

    # make copy of samples.csv to add to
    print("Copying samples csv...")
    shutil.copyfile(path_to_samples, cwd + "/" + out_file_string + ".csv")

    # make new columns in csv for the alignment stats
    alignment_stats = pd.read_csv(cwd + "/" + out_file_string + ".csv",index_col=False) # index_col=False needed so that Path does not become row name
    #alignment_stats.iloc[0]
    alignment_stats = alignment_stats.astype(str)
    alignment_stats["Number Aligned Once"] = ""
    alignment_stats["Percent Aligned Once"] = ""
    alignment_stats["Percent Overall Alignment"] = ""

    # grab relevant info from log files
    os.system("grep exactly 1-Mapping/bowtie2/bowtie2_*_ref_" + reference_genome + ".txt" " >> " + cwd + "/" + out_file_string + ".txt")
    os.system("grep overall 1-Mapping/bowtie2/bowtie2_*_ref_" + reference_genome + ".txt" " >> " + cwd + "/" + out_file_string + ".txt") 

    sample_id_to_number_once  = {}        
    sample_id_to_percent_once = {}
    sample_id_to_overall      = {}

    stats_file = open(cwd + "/" + out_file_string + ".txt")
    for line in stats_file:
        split_on_rule_name = line.split("bowtie2_")
        split_on_ref = split_on_rule_name[1].split("_ref_")
        sample_id = split_on_ref[0]
        split_on_colon = split_on_ref[1].split(":")
        stat_info = split_on_colon[1]
        if "exactly" in stat_info:
            info = stat_info.split()
            number = info[0]
            percent = info[1].replace("(", "").replace(")","").replace("%", "")
            sample_id_to_number_once[sample_id] = number
            sample_id_to_percent_once[sample_id] = percent
        elif "overall" in stat_info:
            info = stat_info.split()
            percent = info[0].replace("%", "")  
            sample_id_to_overall[sample_id] = percent

    for idx, row in alignment_stats.iterrows():
        sample_id_from_csv = row["Sample"]
        ref_id_from_csv = row["Reference"]
        if reference_genome in ref_id_from_csv:
            if sample_id_from_csv in sample_id_to_number_once:
                row["Number Aligned Once"] = sample_id_to_number_once[sample_id_from_csv]
                row["Percent Aligned Once"] = sample_id_to_percent_once[sample_id_from_csv]
                row["Percent Overall Alignment"] = sample_id_to_overall[sample_id_from_csv]
            else:
                row["Number Aligned Once"] = "-1"
                row["Percent Aligned Once"] = "-1"
                row["Percent Overall Alignment"] = "-1"
    # Remove rows where samples weren't aligned to this reference genome
    alignment_stats = alignment_stats[(alignment_stats["Number Aligned Once"] != "")]


    # for each stat, make a histogram across the samples
    number_samples = alignment_stats.shape[0]
    number_bins = math.ceil(math.sqrt(number_samples))
    # number aligned once histogram
    number_aligned_once = alignment_stats["Number Aligned Once"].tolist()
    number_aligned_once = [int(number) for number in number_aligned_once]
    number_aligned_once = np.array(number_aligned_once)
    hist, bin_edges = np.histogram(number_aligned_once, bins=number_bins)
    plt.hist(number_aligned_once, bins=bin_edges, edgecolor='white', linewidth=1.0)
    plt.title("Number of Paired Reads That Aligned Once")
    plt.xlabel("Number of Paired Reads")
    plt.ylabel("Number of Samples")
    plt.savefig(cwd + "/" + out_file_string + "_number_aligned_once_histogram.png")

    # percent aligned once histogram
    percent_aligned_once = alignment_stats["Percent Aligned Once"].tolist()
    percent_aligned_once = [float(percent) for percent in percent_aligned_once]
    percent_aligned_once = np.array(percent_aligned_once)
    hist, bin_edges = np.histogram(percent_aligned_once, bins=number_bins)
    plt.clf()
    plt.hist(percent_aligned_once, bins=bin_edges, edgecolor='white', linewidth=1.0)
    plt.title("Percent of Paired Reads That Aligned Once")
    plt.xlabel("Percent of Paired Reads")
    plt.ylabel("Number of Samples")
    plt.savefig(cwd + "/" + out_file_string + "_percent_aligned_once_histogram.png")

    # overall alignment histogram
    overall_alignment = alignment_stats["Percent Overall Alignment"].tolist()
    overall_alignment = [float(percent) for percent in overall_alignment]
    overall_alignment = np.array(overall_alignment)
    hist, bin_edges = np.histogram(overall_alignment, bins=number_bins)
    plt.clf()
    plt.hist(overall_alignment, bins=bin_edges, edgecolor='white', linewidth=1.0)
    plt.title("Overall Alignment Percent")
    plt.xlabel("Percent of Paired Reads")
    plt.ylabel("Number of Samples")
    plt.savefig(cwd + "/" + out_file_string + "_overall_alignment_histogram.png")

    # save the csv file in the bowtie2 step folder
    path_to_save_csv = cwd + "/" + out_file_string + ".csv" 
    alignment_stats.to_csv(path_to_save_csv) # alignment_stats.to_csv(path_to_save_csv, index = False)
    print("Done with bowtie2 QC")

if __name__ == "__main__":
    path_to_samples=args.samples
    reference_genome=args.refGenome
    current_directory=args.currentDirectory
    out_file_string=args.outfileString
    main(path_to_samples, reference_genome, current_directory, out_file_string)
